import numpy as np
cimport numpy as np
cimport cython
from scipy.optimize import newton
#from libc.math cimport fabs

# suspect that the function _brentq is in c and thus this is the most effective
# method for using the brentq method in cython.
from scipy.optimize._zeros import _brentq as brentq

DTYPE_FLOAT = np.double
ctypedef np.double_t DTYPE_FLOAT_t

DTYPE_INT = np.int
ctypedef np.int_t DTYPE_INT_t


cdef extern from "math.h":
    double fabs(double x) nogil
    double pow(double x, double y) nogil
    
@cython.boundscheck(False)
def erode_avoiding_pits(np.ndarray[DTYPE_INT_t, ndim=1] src_nodes,
                        np.ndarray[DTYPE_INT_t, ndim=1] dst_nodes,
                        np.ndarray[DTYPE_FLOAT_t, ndim=1] node_z,
                        np.ndarray[DTYPE_FLOAT_t, ndim=1] node_dz):
    """Erode node elevations while avoiding creating pits.

    Parameters
    ----------
    src_nodes : array_like
        Ordered upstream node ids.
    dst_nodes : array_like
        Node ids of nodes receiving flow.
    node_z : array_like
        Node elevations.
    node_dz : array_like
        Node erosion rates.
    """
    cdef unsigned int n_nodes = src_nodes.size
    cdef double z_src_after
    cdef double z_dst_after
    cdef unsigned int src_id
    cdef unsigned int dst_id
    cdef unsigned int i

    for i in range(n_nodes):
        src_id = src_nodes[i]
        dst_id = dst_nodes[src_id]

        z_src_after = node_z[src_id] - node_dz[src_id]
        z_dst_after = node_z[dst_id] - node_dz[dst_id]

        if z_src_after < z_dst_after:
            node_dz[src_id] = (node_z[src_id] - z_dst_after) * 0.999999


def erode_with_link_alpha_varthresh(np.ndarray[DTYPE_INT_t, ndim=1] src_nodes,
                                    np.ndarray[DTYPE_INT_t, ndim=1] dst_nodes,
                                    np.ndarray[DTYPE_FLOAT_t, ndim=1] threshsxdt,
                                    np.ndarray[DTYPE_FLOAT_t, ndim=1] alpha,
                                    DTYPE_FLOAT_t n,
                                    np.ndarray[DTYPE_FLOAT_t, ndim=1] z):
    """Erode node elevations using alpha scaled by link length.

    Parameters
    ----------
    src_nodes : array_like
        Ordered upstream node ids.
    dst_nodes : array_like
        Node ids of nodes receiving flow.
    threshsxdt : array_like
        Incision thresholds at nodes multiplied by the timestep.
    alpha : array_like
        Erosion factor scaled by link length to the *n - 1*.
    n : float
        Exponent.
    z : array_like
        Node elevations.
    """
    cdef unsigned int n_nodes = src_nodes.size
    cdef unsigned int src_id
    cdef unsigned int dst_id
    cdef unsigned int i
    cdef double threshxdt
    cdef double z_diff
    cdef double prev_z
    cdef double next_z
    cdef double f
    cdef double excess_thresh

    for i in range(n_nodes):
        src_id = src_nodes[i]
        dst_id = dst_nodes[src_id]
        threshxdt = threshsxdt[i]

        if src_id != dst_id and z[src_id] > z[dst_id]:
            next_z = z[src_id]
            prev_z = 0.
            niter = 0

            while True:
                niter += 1
                if niter > 40:
                    print('===============')
                    print(z[src_id])
                    print(z[dst_id])
                    print(next_z)
                    print(z_diff)
                    print(f)
                    print(alpha[src_id])
                assert niter < 50, 'failure to converge in SP solver'
                z_diff = next_z - z[dst_id]
                f = alpha[src_id] * pow(z_diff, n - 1.)
                excess_thresh = f * z_diff - threshxdt
                if excess_thresh < 0.:
                    excess_thresh = 0.
                next_z = next_z - ((next_z - z[src_id] + excess_thresh) /
                                   (1. + n * f))
                if next_z < z[dst_id]:
                    next_z = z[dst_id] + 1.e-15  # maintain connectivity
                if next_z != 0.:
                    if fabs((next_z - prev_z)/next_z) < 1.48e-08 or n == 1.:
                        break
                else:
                    break

                prev_z = next_z;

            if next_z < z[src_id]:
                z[src_id] = next_z


def erode_with_link_alpha_fixthresh(np.ndarray[DTYPE_INT_t, ndim=1] src_nodes,
                                    np.ndarray[DTYPE_INT_t, ndim=1] dst_nodes,
                                    DTYPE_FLOAT_t threshxdt,
                                    np.ndarray[DTYPE_FLOAT_t, ndim=1] alpha,
                                    DTYPE_FLOAT_t n,
                                    np.ndarray[DTYPE_FLOAT_t, ndim=1] z):
    """Erode node elevations using alpha scaled by link length.

    Parameters
    ----------
    src_nodes : array_like
        Ordered upstream node ids.
    dst_nodes : array_like
        Node ids of nodes receiving flow.
    threshxdt : float
        The (spatially uniform) incision threshold multiplied by the timestep.
    alpha : array_like
        Erosion factor scaled by link length to the *n - 1*.
    n : float
        Exponent.
    z : array_like
        Node elevations.
    """
    cdef unsigned int n_nodes = src_nodes.size
    cdef unsigned int src_id
    cdef unsigned int dst_id
    cdef unsigned int i
    cdef unsigned int niter
    cdef double z_diff
    cdef double prev_z
    cdef double next_z
    cdef double f
    cdef double excess_thresh

    for i in range(n_nodes):
        src_id = src_nodes[i]
        dst_id = dst_nodes[src_id]

        if src_id != dst_id and z[src_id] > z[dst_id]:
            next_z = z[src_id]
            prev_z = 0.
            niter = 0

            while True:
                niter += 1
                if niter > 40:
                    print('===============')
                    print(z[src_id])
                    print(z[dst_id])
                    print(next_z)
                    print(z_diff)
                    print(f)
                    print(alpha[src_id])
                assert niter < 50, 'failure to converge in SP solver'
                z_diff = next_z - z[dst_id]
                f = alpha[src_id] * pow(z_diff, n - 1.)
                excess_thresh = f * z_diff - threshxdt
                if excess_thresh < 0.:
                    excess_thresh = 0.
                next_z = next_z - ((next_z - z[src_id] + excess_thresh) /
                                   (1. + n * f))
                if next_z < z[dst_id]:
                   next_z = z[dst_id] + 1.e-15  # maintain connectivity
                if next_z != 0.:
                    if fabs((next_z - prev_z)/next_z) < 1.48e-08 or n == 1.:
                        break
                else:
                    break

                prev_z = next_z;

            if next_z < z[src_id]:
                z[src_id] = next_z


def brent_method_erode_variable_threshold(np.ndarray[DTYPE_INT_t, ndim=1] src_nodes,
                                          np.ndarray[DTYPE_INT_t, ndim=1] dst_nodes,
                                          np.ndarray[DTYPE_FLOAT_t, ndim=1] threshsxdt,
                                          np.ndarray[DTYPE_FLOAT_t, ndim=1] alpha,
                                          DTYPE_FLOAT_t n,
                                          np.ndarray[DTYPE_FLOAT_t, ndim=1] z):
    """Erode node elevations using Brent's method for stability.

    The alpha value is given as

    alpha = delta_t*K * (rainfall_intensity*A)**m/(delta_x**n)

    It will be multiplied by the value:
        (z_node(t) - z_downstream(t+delta_t))**(n-1)

    to become the alpha value defined below in the function used in erode_fun
    used in the root-finding operation.

    Parameters
    ----------
    src_nodes : array_like
        Ordered upstream node ids.
    dst_nodes : array_like
        Node ids of nodes receiving flow.
    threshsxdt : array_like
        Incision thresholds at nodes multiplied by the timestep.
    alpha : array_like
        Erosion factor.
    n : float
        Exponent.
    z : array_like
        Node elevations.
    """
    # define internally used variables.
    cdef unsigned int n_nodes = src_nodes.size
    cdef unsigned int src_id
    cdef unsigned int dst_id
    cdef unsigned int i
    cdef double z_old
    cdef double z_downstream
    cdef double thresholddt
    cdef double z_diff_old
    cdef double z_diff
    cdef double alpha_param
    cdef double beta_param
    cdef double check_function
    cdef double x

    # Loop through nodes.
    for i in range(n_nodes):

        # get IDs for source and reciever nodes
        src_id = src_nodes[i]
        dst_id = dst_nodes[src_id]

        # if a node does not flow to itself, and the source node is above the
        # destination node
        if src_id != dst_id and z[src_id] > z[dst_id]:

            # Get values for z at present node and present time,
            # and z downstream at t + delta t (which should have been
            # previously solved for)
            z_old = z[src_id]
            z_downstream = z[dst_id]

            # Get the threshold value. In this function, it is spatially variable
            thresholddt = threshsxdt[src_id]

            # calculate the difference between z_old and z_downstream
            z_diff_old = z_old - z_downstream

            # using z_diff_old, calculate the alpha paramter of Braun and
            # Willet by calculating alpha times z

            alpha_param = alpha[src_id] * pow(z_diff_old, n-1.0)

            # Calculate the beta parameter that accounts for the possible
            # presence of a threshold.
            beta_param = thresholddt / z_diff_old

            # check if the threshold has been exceeded by passing a value of
            # x = 1 to the erode_fn. If this returns a value of less than
            # zero, this means that the the maximum possible slope value  does
            # not produce stream power needed to exceed the erosion threshold
            check_function = erode_fn(1, alpha_param, beta_param, n)

            # if the threshold was not exceeded do not change the elevation,
            # otherwise calculate the erosion rate
            if check_function > 0:
                # if the threshold was exceeded, then there will be a zero
                # between x = 0 and x= 1

                # solve using brentq, which requires a zero to exist
                # in between the two end values

                # if n is 1, finding x has an analytical solution. Otherwise,
                # use the the numerical solution given by root finding
                if n != 1.0:

                    # The threshold values passed here are the defaults if one
                    # were to import brentq from scipy.optimize
                    x = brentq(erode_fn,
                                       0.0, 1.0,
                                       1e-12,4.4408920985006262e-16,
                                       100,
                                       (alpha_param, beta_param, n),
                                       False,
                                       True)

                else:
                    # Analytical solution
                    x = (1.0 + beta_param)/(1.0 + alpha_param)

                # If x is provided as a value greater than zero, calculate
                # z at t=t+delta_t useing the values of x, z_downstream and
                # z_old as given by the definition of x (see erode_fn for
                # details). If x is equal to zero, set it as just slightly
                # higher than x_downstream.
                if x>0:
                    z[src_id] = z_downstream + x * (z_old - z_downstream)
                else:
                    z[src_id] = z_downstream + 1.0e-15

                # Nothing is returned from this function as it serves to update
                # the array z.


def brent_method_erode_fixed_threshold(np.ndarray[DTYPE_INT_t, ndim=1] src_nodes,
                                       np.ndarray[DTYPE_INT_t, ndim=1] dst_nodes,
                                       DTYPE_FLOAT_t threshsxdt,
                                       np.ndarray[DTYPE_FLOAT_t, ndim=1] alpha,
                                       DTYPE_FLOAT_t n,
                                       np.ndarray[DTYPE_FLOAT_t, ndim=1] z):
    """Erode node elevations.

    The alpha value is given as

    alpha = delta_t*K * (rainfall_intensity*A)**m/(delta_x**n)

    It will be multiplied by the value:
        (z_node(t) - z_downstream(t+delta_t))**(n-1)

    to become the alpha value defined below in the function used in erode_fun
    used in the root-finding operation.

    Parameters
    ----------
    src_nodes : array_like
        Ordered upstream node ids.
    dst_nodes : array_like
        Node ids of nodes receiving flow.
    threshsxdt : float
        Incision thresholds at nodes multiplied by the timestep.
    alpha : array_like
        Erosion factor.
    n : float
        Exponent.
    z : array_like
        Node elevations.
    """
    # define internally used variables.
    cdef unsigned int n_nodes = src_nodes.size
    cdef unsigned int src_id
    cdef unsigned int dst_id
    cdef unsigned int i
    cdef double z_old
    cdef double z_downstream
    cdef double z_diff_old
    cdef double z_diff
    cdef double alpha_param
    cdef double beta_param
    cdef double check_function
    cdef double x

    # Loop through nodes.
    for i in range(n_nodes):

        # get IDs for source and reciever nodes
        src_id = src_nodes[i]
        dst_id = dst_nodes[src_id]

        # if a node does not flow to itself, and the source node is above the
        # destination node
        if src_id != dst_id and z[src_id] > z[dst_id]:

            # Get values for z at present node and present time,
            # and z downstream at t + delta t (which should have been
            # previously solved for)
            z_old = z[src_id]
            z_downstream = z[dst_id]

            # Get the threshold value. In this function, it is constant, so we
            # already have it.
            # threshsxdt = threshsxdt

            # calculate the difference between z_old and z_downstream
            z_diff_old = z_old - z_downstream

            # using z_diff_old, calculate the alpha paramter of Braun and
            # Willet by calculating alpha times z

            alpha_param = alpha[src_id] * pow(z_diff_old, n-1.0)

            # Calculate the beta parameter that accounts for the possible
            # presence of a threshold.
            beta_param = threshsxdt / z_diff_old
            # check if the threshold has been exceeded by passing a value of
            # x = 1 to the erode_fn. If this returns a value of less than
            # zero, this means that the the maximum possible slope value  does
            # not produce stream power needed to exceed the erosion threshold
            check_function = erode_fn(1, alpha_param, beta_param, n)

            # if the threshold was not exceeded do not change the elevation,
            # otherwise calculate the erosion rate
            if check_function > 0:
                # if the threshold was exceeded, then there will be a zero
                # between x = 0 and x= 1

                # solve using brentq, which requires a zero to exist
                # in between the two end values

                # if n is 1, finding x has an analytical solution. Otherwise,
                # use the the numerical solution given by root finding
                if n != 1.0:

                    # The threshold values passed here are the defaults if one
                    # were to import brentq from scipy.optimize
                    x = brentq(erode_fn,
                                       0.0, 1.0,
                                       1e-12,4.4408920985006262e-16,
                                       100,
                                       (alpha_param, beta_param, n),
                                       False,
                                       True)

                else:
                    # Analytical solution
                    x = (1.0 + beta_param)/(1.0 + alpha_param)

                # If x is provided as a value greater than zero, calculate
                # z at t=t+delta_t useing the values of x, z_downstream and
                # z_old as given by the definition of x (see erode_fn for
                # details). If x is equal to zero, set it as just slightly
                # higher than x_downstream.
                if x>0:
                    z[src_id] = z_downstream + x * (z_old - z_downstream)
                else:
                    z[src_id] = z_downstream + 1.0e-15

                # Nothing is returned from this function as it serves to update
                # the array z.


def erode_fn(DTYPE_FLOAT_t x,
             DTYPE_FLOAT_t alpha,
             DTYPE_FLOAT_t beta,
             DTYPE_FLOAT_t n):
    """Evaluates the solution to the water-depth equation.

    Called by scipy.brentq() to find solution for $x$ using Brent's method.

    Parameters
    ----------
    x : float
        normalized elevation, see below.
    alpha : float
        alpha parameter, see below.
    beta : float
        beta parameter, see below.
    n : float
        n exponent
    check : boolean, default is True
        flag to determine if a ValueError should be thrown if a check
        identifies that the threshold value is high enough such that no erosion
        will occur.


    This equation represents the implicit solution for normalized topographic
    elevation $x$ at the  next time step. This solution is inspired by the
    Appendix of Braun and Willet (2012) but was generalized to include an a
    threshold value such that if the threshold is not exceeded, no erosion will
    occur.

    Consider stream power erosion under the equation:

        E = K * (rainfall_intensity*A)**m * S**n - threshold_sp,

    on a grid with link delta_x and for a timestep of delta_t.

    When iterating from downstream to upstream in the drainage stack, at a
    given node at time = t+delta_t, the value of the node at time = t, and the
    value of the downstream node at time t+delta_t is known.

    Define
    x = (z_node(t+delta_t) - z_downstream(t+delta_t))/(z_node(t) - z_downstream(t+delta_t)).

    A discretized version of the stream power equation above yeilds the equation

    f = x - 1 + alpha*(x**n) - beta

    where

    alpha = delta_t*K * (rainfall_intensity*A)**m/(delta_x**n) * (z_node(t) - z_downstream(t+delta_t))**(n-1)

    and

    beta = threshold_sp*delta_t/(z_node(t) - z_downstream(t+delta_t))

    Finding the root of f provides the implicit solution for the stream power
    equation.

    If f(x=1) = 0, then no erosion occurs as potential erosion is cancelled by
    the erosion threshold and the topography at the given node does not change.

    If f(x=0) = 0, then the topography at the given node becomes that of the
    the downstream mode.

    If the threshold term, beta, is zero, this equation collapses to the form
    given by Braun and Willet (2012).

    When the threshold term is greater than zero, it is possible that no
    erosion will occur. In this case, an evaluation of f(x=1) will yeild a
    negative number.

    If for the values of alpha and beta provided, f(x=1)<0, then this function
    has no root and use in a solver such as Brent's method in which a solution
    interval is required will fail.

    It is recommended that before using this method in a solver, that the
    function be evaluated with x=1 to determine if it any erosion occured.

    """
    cdef double f

    f = x - 1.0 + (alpha * (x ** n)) - beta

    return f


def smooth_stream_power_eroder_solver(np.ndarray[DTYPE_INT_t, ndim=1] src_nodes,
                                      np.ndarray[DTYPE_INT_t, ndim=1] dst_nodes,
                                      np.ndarray[DTYPE_FLOAT_t, ndim=1] z,
                                      np.ndarray[DTYPE_FLOAT_t, ndim=1] alpha,
                                      np.ndarray[DTYPE_FLOAT_t, ndim=1] gamma,
                                      np.ndarray[DTYPE_FLOAT_t, ndim=1] delta):
    """Erode node elevations using Newtons Method for smoothed Stream Power. "

    This method takes three parameters, alpha, gamma, and delta. 

    alpha = K A^m dt / L
    
    delta = K A^m / (L * wc)
    
    gamma = omega_c * dt 

    This method will use the new_elev and new_elev_prime equations. 

    Parameters
    ----------
    src_nodes : array_like
        Ordered upstream node ids.
    dst_nodes : array_like
        Node ids of nodes receiving flow.
    alpha : array_like
        Erosion equation parameter. 
    gamma : array_like
        Erosion equation parameter. 
    delta : array_like
        Erosion equation parameter. 
    z : array_like
        Node elevations.
    """
    cdef unsigned int n_nodes = src_nodes.size
    cdef unsigned int src_id
    cdef unsigned int dst_id
    cdef unsigned int i

    cdef double epilon



    for i in range(len(src_nodes)):
        src_id = src_nodes[i]
        dst_id = dst_nodes[src_id]

        if src_id != dst_id and z[src_id] > z[dst_id]:

            # calculate epsilon
            epsilon = (alpha[src_id] * z[dst_id]
                       + gamma[src_id] + z[src_id])

            # calculate new z
            z[src_id] = newton(new_elev, z[src_id],
                             fprime=new_elev_prime,
                             args=(alpha[src_id],
                                   z[dst_id],
                                   gamma[src_id],
                                   delta[src_id],
                                   epsilon))


def new_elev(DTYPE_FLOAT_t x,
             DTYPE_FLOAT_t a,
             DTYPE_FLOAT_t b,
             DTYPE_FLOAT_t c,
             DTYPE_FLOAT_t d,
             DTYPE_FLOAT_t e):
    """Equation for elevation of a node at timestep t+1.

    Parameters
    ----------
    x : float
        Value of new elevation
    a : float
        Parameter = K A^m dt / L (nondimensional)
    b : float
        Elevation of downstream node, z_j
    c : float
        Parameter = omega_c * dt (dimension of L, because omega_c [=] L/T)
    d : float
        Parameter = K A^m / (L * wc) [=] L^{-1} (so d * z [=] [-])
    e : float
        z(t) + a z_j + (wc * dt)
    """
    cdef double f

    f = x * (1.0 + a) + c * np.exp(-d * (x - b)) - e

    return f


def new_elev_prime(DTYPE_FLOAT_t x,
                   DTYPE_FLOAT_t a,
                   DTYPE_FLOAT_t b,
                   DTYPE_FLOAT_t c,
                   DTYPE_FLOAT_t d,
                   DTYPE_FLOAT_t e):
    """Equation for elevation of a node at timestep t+1.

    Parameters
    ----------
    x : float
        Value of new elevation
    a : float
        Parameter = K A^m dt / L
    b : float
        Elevation of downstream node, z_j
    c : float
        Parameter = omega_c * dt (dimension of L, because omega_c [=] L/T)
    d : float
        Parameter = K A^m / (L * wc) [=] L^{-1} (so d * z [=] [-])
    e : n/a
        Placeholder; not used
    """
    cdef double f

    f = (1.0 + a) - c * d * np.exp(-d * (x - b))

    return f


import numpy as np
cimport numpy as np
cimport cython

from libc.stdlib cimport malloc, free, qsort
from libc.math cimport atan2


DTYPE = np.int
ctypedef np.int_t DTYPE_t

from libc.stdlib cimport malloc, free, qsort


ctypedef np.int_t INT_t
ctypedef np.float_t FLOAT_t

cdef struct Sorter:
    INT_t index
    FLOAT_t value


cdef extern from "stdlib.h":
    ctypedef void const_void "const void"
    void qsort(void *base, int nmemb, int size,
            int(*compar)(const_void *, const_void *)) nogil


cdef int _compare(const_void *a, const_void *b):
    cdef double v = ((<Sorter*>a)).value - ((<Sorter*>b)).value
    if v < 0:
        return -1
    else:
        return 1


cdef void _argsort(double * data, int n_elements, Sorter * order):
    cdef int i

    for i in range(n_elements):
        order[i].index = i
        order[i].value = data[i]

    qsort(<void*> order, n_elements, sizeof(Sorter), _compare)


cdef argsort(double * data, int n_elements, int * out):
    cdef Sorter *sorted_struct = <Sorter*>malloc(n_elements * sizeof(Sorter))

    try:
        _argsort(data, n_elements, sorted_struct)
        for i in range(n_elements):
            out[i] = sorted_struct[i].index
    finally:
        free(sorted_struct)


cdef argsort_inplace(double * data, int n_elements, int * out):
    cdef Sorter *sorted_struct = <Sorter*>malloc(n_elements * sizeof(Sorter))

    try:
        _argsort(data, n_elements, sorted_struct)
        for i in range(n_elements):
            out[i] = sorted_struct[i].index
            data[i] = sorted_struct[i].value
    finally:
        free(sorted_struct)


@cython.boundscheck(False)
def _find_links_at_node(DTYPE_t node,
                        np.ndarray[DTYPE_t, ndim=2] nodes_at_link,
                        np.ndarray[DTYPE_t, ndim=1] links_at_node,
                        np.ndarray[DTYPE_t, ndim=1] link_dirs_at_node):
  """Find directions of links touching a node.

  Parameters
  ----------
  node : int
      A node ID.
  nodes_at_link : ndarray of int, shape `(n_links, 2)`
      Nodes at link tail and head.
  links_at_node : ndarray of int, shape `(max_links_per_node, )`
      Buffer to hold link IDs for links around node.
  link_dirs_at_node : ndarray of int, shape `(max_links_per_node, )`
      Buffer to hold link directions for links around node.

  Returns
  -------
  int
    The number of links found.
  """
  cdef int link = 0
  cdef int n_links_found = 0
  cdef int max_links_at_node = links_at_node.shape[0]
  cdef int n_links = nodes_at_link.shape[0]

  while n_links_found < max_links_at_node and link < n_links:
    if nodes_at_link[link, 0] == node:
      links_at_node[n_links_found] = link
      link_dirs_at_node[n_links_found] = -1
      n_links_found += 1
    elif nodes_at_link[link, 1] == node:
      links_at_node[n_links_found] = link
      link_dirs_at_node[n_links_found] = 1
      n_links_found += 1

    link += 1

  return n_links_found


@cython.boundscheck(False)
def _setup_links_at_node(np.ndarray[DTYPE_t, ndim=2] nodes_at_link,
                         np.ndarray[DTYPE_t, ndim=2] links_at_node,
                         np.ndarray[DTYPE_t, ndim=2] link_dirs_at_node):
  cdef int node
  cdef int n_nodes

  n_nodes = links_at_node.shape[0]

  for node in range(n_nodes):
    _find_links_at_node(node, nodes_at_link, links_at_node[node],
                        link_dirs_at_node[node])


@cython.boundscheck(False)
def _setup_node_at_cell(shape,
                        np.ndarray[DTYPE_t, ndim=1] node_at_cell):
  cdef int cell
  cdef int cell_rows = shape[0] - 2
  cdef int cell_cols = shape[1] - 2
  cdef int node_cols = shape[1]
  cdef int row_offset
  cdef int row
  cdef int col


  cell = 0
  row_offset = shape[1] + 1
  for row in range(cell_rows):
    for col in range(cell_cols):
      node_at_cell[cell] = row_offset + col
      cell += 1
    row_offset += node_cols


@cython.boundscheck(False)
def _remap_nodes_at_link(np.ndarray[DTYPE_t, ndim=2] nodes_at_link,
                         np.ndarray[DTYPE_t, ndim=1] new_nodes):
  cdef int link
  cdef int n_links = len(nodes_at_link)

  for link in range(n_links):
    nodes_at_link[link, 0] = new_nodes[nodes_at_link[link, 0]]
    nodes_at_link[link, 1] = new_nodes[nodes_at_link[link, 1]]


@cython.boundscheck(False)
def _remap_links_at_patch(np.ndarray[DTYPE_t, ndim=1] links_at_patch,
                          np.ndarray[DTYPE_t, ndim=1] new_links):
    cdef int link
    cdef int n_links = links_at_patch.shape[0]

    for link in range(n_links):
        links_at_patch[link] = new_links[links_at_patch[link]]


@cython.boundscheck(False)
def _calc_center_of_patch(np.ndarray[DTYPE_t, ndim=1] links_at_patch,
                          np.ndarray[DTYPE_t, ndim=1] offset_to_patch,
                          np.ndarray[np.float_t, ndim=2] xy_at_link,
                          np.ndarray[np.float_t, ndim=2] xy_at_patch):
    cdef int patch
    cdef int link
    cdef int i
    cdef int offset
    cdef int n_links
    cdef int n_patches = len(xy_at_patch)
    cdef float x
    cdef float y

    for patch in range(n_patches):
        offset = offset_to_patch[patch]
        n_links = offset_to_patch[patch + 1] - offset
        x = 0.
        y = 0.
        for i in range(offset, offset + n_links):
            link = links_at_patch[i]
            x += xy_at_link[link, 0]
            y += xy_at_link[link, 1]
        xy_at_patch[patch, 0] = x / n_links
        xy_at_patch[patch, 1] = y / n_links


@cython.boundscheck(False)
def _resort_patches(np.ndarray[DTYPE_t, ndim=1] links_at_patch,
                    np.ndarray[DTYPE_t, ndim=1] offset_to_patch,
                    np.ndarray[DTYPE_t, ndim=1] sorted_patches):
    cdef int i
    cdef int patch
    cdef int offset
    cdef int n_links
    cdef int n_patches = len(sorted_patches)
    cdef int *new_offset = <int *>malloc(len(offset_to_patch) * sizeof(int))
    cdef int *new_patches = <int *>malloc(len(links_at_patch) * sizeof(int))

    try:
        new_offset[0] = 0
        for patch in range(n_patches):
            offset = offset_to_patch[sorted_patches[patch]]
            n_links = offset_to_patch[sorted_patches[patch] + 1] - offset

            new_offset[patch + 1] = new_offset[patch] + n_links
            for i in range(n_links):
                new_patches[new_offset[patch] + i] = links_at_patch[offset + i]

        for i in range(len(links_at_patch)):
            links_at_patch[i] = new_patches[i]
        for i in range(len(offset_to_patch)):
            offset_to_patch[i] = new_offset[i]
    finally:
        free(new_offset)
        free(new_patches)


@cython.boundscheck(False)
def _setup_links_at_patch(np.ndarray[DTYPE_t, ndim=1] links_at_patch,
                          np.ndarray[DTYPE_t, ndim=1] offset_to_patch,
                          np.ndarray[DTYPE_t, ndim=2] out):
    cdef int i
    cdef int link
    cdef int patch
    cdef int offset
    cdef int n_links
    cdef int n_patches = len(offset_to_patch) - 1

    for patch in range(n_patches):
        offset = offset_to_patch[patch]
        n_links = offset_to_patch[patch + 1] - offset

        link = 0
        for i in range(offset, offset + n_links):
          out[patch, link] = links_at_patch[i]
          link += 1


@cython.boundscheck(False)
def _reorder_links_at_node(np.ndarray[DTYPE_t, ndim=2] links_at_node,
                           np.ndarray[DTYPE_t, ndim=2] sorted_links):
    cdef int n_nodes = links_at_node.shape[0]
    cdef int n_links_per_node = links_at_node.shape[1]
    cdef int i
    cdef int node
    cdef int *buffer = <int *>malloc(n_links_per_node * sizeof(int))

    try:
      for node in range(n_nodes):
          for i in range(n_links_per_node):
              buffer[i] = links_at_node[node, sorted_links[node, i]]
          for i in range(n_links_per_node):
            links_at_node[node, i] = buffer[i]
    finally:
        free(buffer)


cdef calc_centroid(double * points, np.int_t n_points, double * out):
    cdef int i
    cdef double xc = 0.
    cdef double yc = 0.

    for i in range(0, 2 * n_points, 2):
        xc += points[i]
        yc += points[i + 1]
    xc /= n_points
    yc /= n_points

    out[0] = xc
    out[1] = yc


cdef calc_spoke_angles(double * hub, double * spokes, np.int_t n_spokes,
                       double * angles):
    cdef int i
    cdef double x0 = hub[0]
    cdef double y0 = hub[1]
    cdef double * spoke = spokes
    cdef double two_pi = 2. * np.pi

    for i in range(n_spokes):
        x = spoke[0]
        y = spoke[1]

        angles[i] = atan2(y - y0, x - x0)
        if angles[i] < 0.:
            angles[i] += two_pi
        spoke += 2


cdef argsort_by_angle_around_centroid(double * points,
                                      np.int_t n_points,
                                      int * out):
    cdef double *hub = <double *>malloc(2 * sizeof(double))
    cdef double *angles = <double *>malloc(n_points * sizeof(double))

    try:
        calc_centroid(points, n_points, hub)
        calc_spoke_angles(hub, points, n_points, angles)
        argsort(angles, n_points, out)
    finally:
        free(angles)
        free(hub)


cdef sort_spokes_around_hub(int * spokes, int n_spokes, double * xy_of_spoke,
                            double * xy_of_hub):
    cdef int point
    cdef int spoke
    cdef double * points = <double *>malloc(2 * n_spokes * sizeof(double))
    cdef double * angles = <double *>malloc(n_spokes * sizeof(double))
    cdef int * ordered = <int *>malloc(n_spokes * sizeof(int))
    cdef int * temp = <int *>malloc(n_spokes * sizeof(int))
    
    try:
        point = 0
        for spoke in range(n_spokes):
            points[point] = xy_of_spoke[2 * spokes[spoke]]
            points[point + 1] = xy_of_spoke[2 * spokes[spoke] + 1]
            point += 2

        calc_spoke_angles(xy_of_hub, points, n_spokes, angles)
        argsort(angles, n_spokes, ordered)

        for spoke in range(n_spokes):
            temp[spoke] = spokes[ordered[spoke]]

        for spoke in range(n_spokes):
            spokes[spoke] = temp[spoke]
    finally:
        free(angles)
        free(temp)
        free(ordered)
        free(points)


@cython.boundscheck(False)
def reorder_links_at_patch(np.ndarray[DTYPE_t, ndim=1] links_at_patch,
                           np.ndarray[DTYPE_t, ndim=1] offset_to_patch,
                           np.ndarray[np.float_t, ndim=2] xy_of_link):
    cdef int n_links = xy_of_link.shape[0]
    cdef int n_patches = len(offset_to_patch) - 1
    cdef int link
    cdef int patch
    cdef int offset
    cdef int i
    cdef double *angles = <double *>malloc(n_links * sizeof(double))
    cdef double *points = <double *>malloc(2 * n_links * sizeof(double))
    cdef int *ordered = <int *>malloc(n_links * sizeof(int))
    cdef int *link_buffer = <int *>malloc(n_links * sizeof(int))

    try:
      for patch in range(n_patches):
          offset = offset_to_patch[patch]
          n_links = offset_to_patch[patch + 1] - offset

          for i in range(n_links):
              link = links_at_patch[offset + i]
              points[2 * i] = xy_of_link[link][0]
              points[2 * i + 1] = xy_of_link[link][1]

          argsort_by_angle_around_centroid(points, n_links, ordered)

          for i in range(n_links):
              link_buffer[i] = links_at_patch[offset + ordered[i]]

          for i in range(n_links):
              links_at_patch[offset + i] = link_buffer[i]

    finally:
        free(link_buffer)
        free(ordered)
        free(points)
        free(angles)


@cython.boundscheck(False)
def sort_spokes_at_wheel(np.ndarray[int, ndim=1] spokes_at_wheel,
                         np.ndarray[int, ndim=1] offset_to_wheel,
                         np.ndarray[double, ndim=2] xy_of_hub,
                         np.ndarray[double, ndim=2] xy_of_spoke):
    """Sort spokes about multiple hubs.

    Parameters
    ----------
    spokes_at_wheel : ndarray of int
        Spokes for each wheel.
    offset_to_wheel : ndarray of int
        Offset into *spokes_at_wheel* for each wheel.
    xy_of_hub : ndarray of float, shape `(n_hubs, 2)`
        Coordinates of each hub as `(x, y)`.
    xy_of_spoke : ndarray of float, shape `(n_spokes, 2)`
        Coordinates of the end of each spoke as `(x, y)`.
    """
    cdef int n_wheels = len(offset_to_wheel) - 1
    cdef int i
    cdef int n_spokes
    cdef int * wheel

    wheel = &spokes_at_wheel[0]
    for i in range(n_wheels):
        n_spokes = offset_to_wheel[i + 1] - offset_to_wheel[i]

        sort_spokes_around_hub(wheel, n_spokes, &xy_of_spoke[0, 0],
                               &xy_of_hub[i, 0])

        wheel += n_spokes


@cython.boundscheck(False)
def argsort_spoke_angles(np.ndarray[double, ndim=2, mode="c"] points,
                         np.ndarray[double, ndim=1, mode="c"] hub,
                         np.ndarray[int, ndim=1] out):
    """Sort spokes by angle around a hub.

    Parameters
    ----------
    points : ndarray of float, shape `(n_points, 2)`
        Coordinates of points as (*x*, *y*).
    out : ndarray of int, shape `(n_points, )`
        Indices of sorted points.

    Returns
    -------
    ndarray of int, shape `(n_points, )`
        Indices of sorted points.
    """
    cdef int n_points = points.shape[0]
    cdef double *angles = <double *>malloc(n_points * sizeof(double))

    try:
        calc_spoke_angles(&hub[0], &points[0, 0], n_points, angles)
        argsort(angles, n_points, &out[0])
    finally:
      free(angles)

    return out



#!/usr/env/python
"""Test the creation and execution of a CellLab-CTS model.

Tests the creation and execution of a CellLab-CTS model, by creating a
simple two-state CA on a small grid.

Created by Greg Tucker, May 2015
"""
from __future__ import print_function

import time
from landlab import RasterModelGrid
from landlab.ca.celllab_cts import Transition, CAPlotter
from landlab.ca.raster_cts import RasterCTS


def setup_transition_list():
    """
    Creates and returns a list of Transition() objects to represent state
    transitions for a biased random walk, in which the rate of downward
    motion is greater than the rate in the other three directions.

    Returns
    -------
    xn_list : list of Transition objects
        List of objects that encode information about the link-state
        transitions.

    Notes
    -----
    This doesn't represent any particular process, but rather is simply used
    to test the CA code. The transition rules have 0-0 pairs transitioning to
    0-1 or 1-0 pairs (50/50 chance) and thence to 1-1 pairs, at which point
    there are no further transitions.

    The states and transitions are as follows::

        Pair state      Transition to       Process             Rate (cells/s)
        ==========      =============       =======             ==============
        0 (0-0)         1 (0-1)                                 0.5
                        2 (1-0)                                 0.5
        1 (0-1)         3 (1-1)                                 1.0
        2 (1-0)         3 (1-1)                                 1.0
        3 (1-1)         (none)                                  -

    """
    # Create an empty transition list
    xn_list = []

    # Append two transitions to the list.
    # Note that the arguments to the Transition() object constructor are:
    #  - Tuple representing starting pair state
    #    (left/bottom cell, right/top cell, orientation)
    #  - Tuple representing new pair state
    #    (left/bottom cell, right/top cell, orientation)
    #  - Transition rate (cells per time step, in this case 1 sec)
    #  - Name for transition
    xn_list.append(Transition((0, 0, 0), (0, 1, 0), 0.5, ''))
    xn_list.append(Transition((0, 0, 0), (1, 0, 0), 0.5, ''))
    xn_list.append(Transition((0, 1, 0), (1, 1, 0), 1., ''))
    xn_list.append(Transition((1, 0, 0), (1, 1, 0), 1., ''))

    return xn_list


def main():

    # INITIALIZE

    # User-defined parameters
    nr = 5  # number of rows in grid
    nc = 5  # number of columns in grid
    plot_interval = 10.0   # time interval for plotting, sec
    run_duration = 10.0   # duration of run, sec
    report_interval = 10.0  # report interval, in real-time seconds

    # Remember the clock time, and calculate when we next want to report
    # progress.
    current_real_time = time.time()
    next_report = current_real_time + report_interval

    # Create grid
    mg = RasterModelGrid(nr, nc, 1.0)

    # Make the boundaries be walls
    mg.set_closed_boundaries_at_grid_edges(True, True, True, True)

    # Set up the states and pair transitions.
    ns_dict = {0: 'black', 1: 'white'}
    xn_list = setup_transition_list()

    # Create the node-state array and attach it to the grid
    node_state_grid = mg.add_zeros('node', 'node_state_map', dtype=int)

    # For visual display purposes, set all boundary nodes to fluid
    node_state_grid[mg.closed_boundary_nodes] = 0

    # Create the CA model
    ca = RasterCTS(mg, ns_dict, xn_list, node_state_grid)

    # Create a CAPlotter object for handling screen display
    ca_plotter = CAPlotter(ca)

    # Plot the initial grid
    ca_plotter.update_plot()

    # RUN
    current_time = 0.0
    while current_time < run_duration:

        # Once in a while, print out simulation and real time to let the user
        # know that the sim is running ok
        current_real_time = time.time()
        if current_real_time >= next_report:
            print('Current sim time', current_time, '(',
                  100 * current_time / run_duration, '%)')
            next_report = current_real_time + report_interval

        # Run the model forward in time until the next output step
        ca.run(current_time + plot_interval, ca.node_state,
               plot_each_transition=True, plotter=ca_plotter)
        current_time += plot_interval

        # Plot the current grid
        ca_plotter.update_plot()

    # FINALIZE

    # Plot
    ca_plotter.finalize()

    print('ok, here are the keys')
    print(ca.__dict__.keys())


# If user runs this file, activate the main() function
if __name__ == "__main__":
    #import cProfile
    #fname = 'test_profiler_for_little_ca.txt'
    # cProfile.run('print main(); print') #, fname)
    main()



from __future__ import print_function

import warnings

from six.moves import range

import numpy as np
from time import sleep
from landlab import ModelParameterDictionary, CLOSED_BOUNDARY, Component

from landlab.core.model_parameter_dictionary import MissingKeyError
from landlab.field.scalar_data_fields import FieldError
from landlab.grid.base import BAD_INDEX_VALUE
from landlab.utils.decorators import make_return_array_immutable


class SedDepEroder(Component):
    """
    This module implements sediment flux dependent channel incision
    following::

        E = f(Qs, Qc) * ([a stream power-like term] - [an optional threshold]),

    where E is the bed erosion rate, Qs is the volumetric sediment flux
    into a node, and Qc is the volumetric sediment transport capacity at
    that node.

    This component is under active research and development; proceed with its
    use at your own risk.

    The details of the implementation are a function of the two key
    arguments, *sed_dependency_type* and *Qc*. The former controls the
    shape of the sediment dependent response function f(Qs, Qc), the
    latter controls the way in which sediment transport capacities are
    calculated (primarily, whether a full Meyer-Peter Muller approach is
    used, or whether simpler stream-power-like equations can be assumed).
    For Qc, 'power_law' broadly follows the assumptions in Gasparini et
    al. 2006, 2007; 'MPM' broadly follows those in Hobley et al., 2011.
    Note that a convex-up channel can result in many cases assuming MPM,
    unless parameters b and c are carefully tuned.

    If ``Qc == 'power_law'``::

        E  = K_sp * f(Qs, Qc) * A ** m_sp * S ** n_sp;
        Qc = K_t * A ** m_t * S ** n_t

    If ``Qc == 'MPM'``::

        shear_stress = fluid_density * g * depth * S
                     = fluid_density * g * (mannings_n/k_w) ** 0.6 * (
                       k_Q* A ** c_sp) ** (0.6 * (1. - b_sp)) * S ** 0.7,
                       for consistency with MPM

        E = K_sp * f(Qs, Qc) * (shear_stress ** a_sp - [threshold_sp])

        Qc = 8 * C_MPM * sqrt((sed_density-fluid_density)/fluid_density *
             g * D_char**3) * (shields_stress - threshold_shields)**1.5

        shields_stress = shear_stress / (g * (sed_density-fluid_density) *
                         D_char)

    If you choose Qc='MPM', you may provide thresholds for both channel
    incision and shields number, or alternatively set either or both of
    these threshold dynamically. The minimum shear stress can be made
    equivalent to the Shields number using *set_threshold_from_Dchar*,
    for full consistency with the MPM approach (i.e., the threshold
    becomes a function of the characteristic grain size on the bed). The
    Shields threshold itself can also be a weak function of slope if
    *slope_sensitive_threshold*, following Lamb et al. 2008,
    taustar_c = 0.15 * S ** 0.25.

    The component is able to handle flooded nodes, if created by a lake
    filler. It assumes the flow paths found in the fields already reflect
    any lake routing operations, and then requires the optional argument
    *flooded_depths* be passed to the run method. A flooded depression
    acts as a perfect sediment trap, and will be filled sequentially
    from the inflow points towards the outflow points.
    """

    _name = 'SedDepEroder'

    _input_var_names = (
        'topographic__elevation',
        'drainage_area',
        'flow__receiver_node',
        'flow__upstream_node_order',
        'topographic__steepest_slope',
        'flow__link_to_receiver_node'
    )

    _output_var_names = (
        'topographic__elevation',
        'channel__bed_shear_stress',
        'channel_sediment__volumetric_transport_capacity',
        'channel_sediment__volumetric_flux',
        'channel_sediment__relative_flux',
        'channel__discharge',
        'channel__width',  # optional
        'channel__depth',  # optional
    )

    _optional_var_names = (
        'channel__width',
        'channel__depth'
    )

    _var_units = {'topographic__elevation': 'm',
                  'drainage_area': 'm**2',
                  'flow__receiver_node': '-',
                  'topographic__steepest_slope': '-',
                  'flow__upstream_node_order': '-',
                  'flow__link_to_receiver_node': '-',
                  'channel__bed_shear_stress': 'Pa',
                  'channel_sediment__volumetric_transport_capacity': 'm**3/s',
                  'channel_sediment__volumetric_flux': 'm**3/s',
                  'channel_sediment__relative_flux': '-',
                  'channel__discharge': 'm**3/s',
                  'channel__width': 'm',
                  'channel__depth': 'm'
                  }

    _var_mapping = {'topographic__elevation': 'node',
                    'drainage_area': 'node',
                    'flow__receiver_node': 'node',
                    'topographic__steepest_slope': 'node',
                    'flow__upstream_node_order': 'node',
                    'flow__link_to_receiver_node': 'node',
                    'channel__bed_shear_stress': 'node',
                    'channel_sediment__volumetric_transport_capacity': 'node',
                    'channel_sediment__volumetric_flux': 'node',
                    'channel_sediment__relative_flux': 'node',
                    'channel__discharge': 'node',
                    'channel__width': 'node',
                    'channel__depth': 'node'
                    }

    _var_type = {'topographic__elevation': float,
                 'drainage_area': float,
                 'flow__receiver_node': int,
                 'topographic__steepest_slope': float,
                 'flow__upstream_node_order': int,
                 'flow__link_to_receiver_node': int,
                 'channel__bed_shear_stress': float,
                 'channel_sediment__volumetric_transport_capacity': float,
                 'channel_sediment__volumetric_flux': float,
                 'channel_sediment__relative_flux': float,
                 'channel__discharge': float,
                 'channel__width': float,
                 'channel__depth': float
                 }

    _var_doc = {
        'topographic__elevation': 'Land surface topographic elevation',
        'drainage_area':
            ("Upstream accumulated surface area contributing to the node's " +
             "discharge"),
        'flow__receiver_node':
            ('Node array of receivers (node that receives flow from current ' +
             'node)'),
        'topographic__steepest_slope':
            'Node array of steepest *downhill* slopes',
        'flow__upstream_node_order':
            ('Node array containing downstream-to-upstream ordered list of ' +
             'node IDs'),
        'flow__link_to_receiver_node':
            'ID of link downstream of each node, which carries the discharge',
        'channel__bed_shear_stress':
            ('Shear exerted on the bed of the channel, assuming all ' +
             'discharge travels along a single, self-formed channel'),
        'channel_sediment__volumetric_transport_capacity':
            ('Volumetric transport capacity of a channel carrying all runoff' +
             ' through the node, assuming the Meyer-Peter Muller transport ' +
             'equation'),
        'channel_sediment__volumetric_flux':
            ('Total volumetric fluvial sediment flux brought into the node ' +
             'from upstream'),
        'channel_sediment__relative_flux':
            ('The fluvial_sediment_flux_into_node divided by the fluvial_' +
             'sediment_transport_capacity'),
        'channel__discharge':
            ('Volumetric water flux of the a single channel carrying all ' +
             'runoff through the node'),
        'channel__width':
            ('Width of the a single channel carrying all runoff through the ' +
             'node'),
        'channel__depth':
            ('Depth of the a single channel carrying all runoff through the ' +
             'node')
    }

    def __init__(self, grid, K_sp=1.e-6, g=9.81,
                 rock_density=2700, sediment_density=2700, fluid_density=1000,
                 runoff_rate=1.,
                 sed_dependency_type='generalized_humped', kappa_hump=13.683,
                 nu_hump=1.13, phi_hump=4.24, c_hump=0.00181,
                 Qc='power_law', m_sp=0.5, n_sp=1., K_t=1.e-4, m_t=1.5, n_t=1.,
                 # these params for Qc='MPM':
                 C_MPM=1., a_sp=1., b_sp=0.5, c_sp=1., k_w=2.5, k_Q=2.5e-7,
                 mannings_n=0.05, threshold_shear_stress=None,
                 Dchar=0.05, set_threshold_from_Dchar=True,
                 set_Dchar_from_threshold=False,
                 threshold_Shields=0.05, slope_sensitive_threshold=False,
                 # params for model numeric behavior:
                 pseudoimplicit_repeats=5, return_stream_properties=False,
                 **kwds):
        """Constructor for the class.

        Parameters
        ----------
        grid : a ModelGrid
            A grid.
        K_sp : float (time unit must be *years*)
            K in the stream power equation; the prefactor on the erosion
            equation (units vary with other parameters).
        g : float (m/s**2)
            Acceleration due to gravity.
        rock_density : float (Kg m**-3)
            Bulk intact rock density.
        sediment_density : float (Kg m**-3)
            Typical density of loose sediment on the bed.
        fluid_density : float (Kg m**-3)
            Density of the fluid.
        runoff_rate : float, array or field name (m/s)
            The rate of excess overland flow production at each node (i.e.,
            rainfall rate less infiltration).
        pseudoimplicit_repeats : int
            Number of loops to perform with the pseudoimplicit iterator,
            seeking a stable solution. Convergence is typically rapid.
        return_stream_properties : bool
            Whether to perform a few additional calculations in order to set
            the additional optional output fields, 'channel__width',
            'channel__depth', and 'channel__discharge' (default False).
        sed_dependency_type : {'generalized_humped', 'None', 'linear_decline',
                               'almost_parabolic'}
            The shape of the sediment flux function. For definitions, see
            Hobley et al., 2011. 'None' gives a constant value of 1.
            NB: 'parabolic' is currently not supported, due to numerical
            stability issues at channel heads.
        Qc : {'power_law', 'MPM'}
            Whether to use simple stream-power-like equations for both
            sediment transport capacity and erosion rate, or more complex
            forms based directly on the Meyer-Peter Muller equation and a
            shear stress based erosion model consistent with MPM (per
            Hobley et al., 2011).

        If ``sed_dependency_type == 'generalized_humped'``...

        kappa_hump : float
            Shape parameter for sediment flux function. Primarily controls
            function amplitude (i.e., scales the function to a maximum of 1).
            Default follows Leh valley values from Hobley et al., 2011.
        nu_hump : float
            Shape parameter for sediment flux function. Primarily controls
            rate of rise of the "tools" limb. Default follows Leh valley
            values from Hobley et al., 2011.
        phi_hump : float
            Shape parameter for sediment flux function. Primarily controls
            rate of fall of the "cover" limb. Default follows Leh valley
            values from Hobley et al., 2011.
        c_hump : float
            Shape parameter for sediment flux function. Primarily controls
            degree of function asymmetry. Default follows Leh valley values
            from Hobley et al., 2011.

        If ``Qc == 'power_law'``...

        m_sp : float
            Power on drainage area in the erosion equation.
        n_sp : float
            Power on slope in the erosion equation.
        K_t : float (time unit must be in *years*)
            Prefactor in the transport capacity equation.
        m_t : float
            Power on drainage area in the transport capacity equation.
        n_t : float
            Power on slope in the transport capacity equation.

        if ``Qc == 'MPM'``...

        C_MPM : float
            A prefactor on the MPM relation, allowing tuning to known sediment
            saturation conditions (leave as 1. in most cases).
        a_sp : float
            Power on shear stress to give erosion rate.
        b_sp : float
            Power on drainage area to give channel width.
        c_sp : float
            Power on drainage area to give discharge.
        k_w : float (unit variable with b_sp)
            Prefactor on A**b_sp to give channel width.
        k_Q : float (unit variable with c_sp, but time unit in *seconds*)
            Prefactor on A**c_sp to give discharge.
        mannings_n : float
            Manning's n for the channel.
        threshold_shear_stress : None or float (Pa)
            The threshold shear stress in the equation for erosion rate. If
            None, implies that *set_threshold_from_Dchar* is True, and this
            parameter will get set from the Dchar value and critical Shields
            number.
        Dchar :None, float, array, or field name (m)
            The characteristic grain size on the bed, that controls the
            relationship between critical Shields number and critical shear
            stress. If None, implies that *set_Dchar_from_threshold* is True,
            and this parameter will get set from the threshold_shear_stress
            value and critical Shields number.
        set_threshold_from_Dchar : bool
            If True (default), threshold_shear_stress will be set based on
            Dchar and threshold_Shields.
        set_Dchar_from_threshold : bool
            If True, Dchar will be set based on threshold_shear_stress and
            threshold_Shields. Default is False.
        threshold_Shields : None or float
            The threshold Shields number. If None, implies that
            *slope_sensitive_threshold* is True.
        slope_sensitive_threshold : bool
            If True, the threshold_Shields will be set according to
            0.15 * S ** 0.25, per Lamb et al., 2008 & Hobley et al., 2011.
        """
        self._grid = grid
        self.pseudoimplicit_repeats = pseudoimplicit_repeats

        self.link_S_with_trailing_blank = np.zeros(grid.number_of_links+1)
        # ^needs to be filled with values in execution
        self.count_active_links = np.zeros_like(
            self.link_S_with_trailing_blank, dtype=int)
        self.count_active_links[:-1] = 1

        self._K_unit_time = K_sp/31557600.
        # ^...because we work with dt in seconds
        # set gravity
        self.g = g
        self.rock_density = rock_density
        self.sed_density = sediment_density
        self.fluid_density = fluid_density
        self.relative_weight = (
            (self.sed_density-self.fluid_density)/self.fluid_density*self.g)
        # ^to accelerate MPM calcs
        self.rho_g = self.fluid_density*self.g
        self.type = sed_dependency_type
        assert self.type in ('generalized_humped', 'None', 'linear_decline',
                             'almost_parabolic')
        self.Qc = Qc
        assert self.Qc in ('MPM', 'power_law')
        self.return_ch_props = return_stream_properties
        if return_stream_properties:
            assert self.Qc == 'MPM', ("Qc must be 'MPM' to return stream " +
                                      "properties")
        if type(runoff_rate) in (float, int):
            self.runoff_rate = float(runoff_rate)
        elif type(runoff_rate) is str:
            self.runoff_rate = self.grid.at_node[runoff_rate]
        else:
            self.runoff_rate = np.array(runoff_rate)
            assert runoff_rate.size == self.grid.number_of_nodes

        if self.Qc == 'MPM':
            if threshold_shear_stress is not None:
                self.thresh = threshold_shear_stress
                self.set_threshold = True
                # ^flag for sed_flux_dep_incision to see if the threshold was
                # manually set.
                # print("Found a shear stress threshold to use: ", self.thresh)
            else:
                warnings.warn("Found no incision threshold to use.")
                self.thresh = 0.
                self.set_threshold = False
            self._a = a_sp
            self._b = b_sp
            self._c = c_sp

            self.k_Q = k_Q
            self.k_w = k_w
            self.mannings_n = mannings_n
            if mannings_n < 0. or mannings_n > 0.2:
                warnings.warn("Manning's n outside it's typical range")

            self.diffusivity_power_on_A = 0.9*self._c*(1.-self._b)
            # ^i.e., q/D**(1/6)

            self.override_threshold = set_threshold_from_Dchar
            self.override_Dchar = set_Dchar_from_threshold
            if self.override_threshold:
                assert self.set_threshold is False, (
                    "If set_threshold_from_Dchar, threshold_Shields must be " +
                    "set to None")
                assert self.override_Dchar is False
            if self.override_Dchar:
                assert self.override_threshold is False

            self.shields_crit = threshold_Shields
            self.lamb_flag = slope_sensitive_threshold
            if self.lamb_flag:
                assert self.shields_crit is None, (
                    "If slope_sensitive_threshold, threshold_Shields must " +
                    "be set to None")

        elif self.Qc == 'power_law':
            self._m = m_sp
            self._n = n_sp
            self._Kt = K_t/31557600.  # in sec
            self._mt = m_t
            self._nt = n_t

        # now conditional inputs
        if self.type == 'generalized_humped':
            self.kappa = kappa_hump
            self.nu = nu_hump
            self.phi = phi_hump
            self.c = c_hump

        if self.Qc == 'MPM':
            if Dchar is not None:
                if type(Dchar) in (int, float):
                    self.Dchar_in = float(Dchar)
                elif type(Dchar) is str:
                    self.Dchar_in = self.grid.at_node[Dchar]
                else:
                    self.Dchar_in = np.array(Dchar)
                    assert self.Dchar_in.size == self.grid.number_of_nodes
                assert not self.override_Dchar, (
                    "If set_Dchar_from_threshold, Dchar must be set to None")
            else:
                assert self.override_Dchar
                # remember the threshold getting set is already tau**a
                if not self.lamb_flag:
                    self.Dchar_in = self.thresh/self.g/(
                        self.sed_density-self.fluid_density)/self.shields_crit
                else:
                    self.Dchar_in = None
            self.C_MPM = C_MPM

            if self.override_threshold:
                # print("Overriding any supplied threshold...")
                try:
                    self.thresh = self.shields_crit*self.g*(
                        self.sed_density-self.fluid_density)*self.Dchar_in
                except AttributeError:
                    self.thresh = self.shields_crit*self.g*(
                        self.sed_density-self.fluid_density)*Dchar

            # new 11/12/14
            self.point6onelessb = 0.6*(1.-self._b)
            self.shear_stress_prefactor = self.fluid_density*self.g*(
                self.mannings_n/self.k_w)**0.6

            if self.set_threshold is False or self.override_threshold:
                try:
                    self.shields_prefactor_to_shear = (
                        (self.sed_density-self.fluid_density)*self.g *
                        self.Dchar_in)
                except AttributeError:  # no Dchar
                    self.shields_prefactor_to_shear_noDchar = (
                        self.sed_density-self.fluid_density)*self.g

            twothirds = 2./3.
            self.Qs_prefactor = (
                4.*self.C_MPM**twothirds*self.fluid_density**twothirds /
                (self.sed_density-self.fluid_density)**twothirds*self.g **
                (twothirds/2.)*mannings_n**0.6*self.k_w**(1./15.)*self.k_Q **
                (0.6+self._b/15.)/self.sed_density**twothirds)
            self.Qs_thresh_prefactor = (
                4.*(self.C_MPM*self.k_w*self.k_Q**self._b/self.fluid_density **
                    0.5/(self.sed_density-self.fluid_density)/self.g /
                    self.sed_density)**twothirds)
            # both these are divided by sed density to give a vol flux
            self.Qs_power_onA = self._c*(0.6+self._b/15.)
            self.Qs_power_onAthresh = twothirds*self._b*self._c

        self.cell_areas = np.empty(grid.number_of_nodes)
        self.cell_areas.fill(np.mean(grid.area_of_cell))
        self.cell_areas[grid.node_at_cell] = grid.area_of_cell

        # set up the necessary fields:
        self.initialize_output_fields()
        if self.return_ch_props:
            self.initialize_optional_output_fields()

    def get_sed_flux_function(self, rel_sed_flux):
        if self.type == 'generalized_humped':
            "Returns K*f(qs,qc)"
            sed_flux_fn = self.kappa*(rel_sed_flux**self.nu + self.c)*np.exp(
                -self.phi*rel_sed_flux)
        elif self.type == 'linear_decline':
            sed_flux_fn = (1.-rel_sed_flux)
        elif self.type == 'parabolic':
            raise MissingKeyError(
                'Pure parabolic (where intersect at zero flux is exactly ' +
                'zero) is currently not supported, sorry. Try ' +
                'almost_parabolic instead?')
            sed_flux_fn = 1. - 4.*(rel_sed_flux-0.5)**2.
        elif self.type == 'almost_parabolic':
            sed_flux_fn = np.where(rel_sed_flux > 0.1,
                                   1. - 4.*(rel_sed_flux-0.5)**2.,
                                   2.6*rel_sed_flux+0.1)
        elif self.type == 'None':
            sed_flux_fn = 1.
        else:
            raise MissingKeyError(
                'Provided sed flux sensitivity type in input file was not ' +
                'recognised!')
        return sed_flux_fn

    def get_sed_flux_function_pseudoimplicit(self, sed_in, trans_cap_vol_out,
                                             prefactor_for_volume,
                                             prefactor_for_dz):
        rel_sed_flux_in = sed_in/trans_cap_vol_out
        rel_sed_flux = rel_sed_flux_in

        if self.type == 'generalized_humped':
            "Returns K*f(qs,qc)"

            def sed_flux_fn_gen(rel_sed_flux_in):
                return self.kappa*(rel_sed_flux_in**self.nu + self.c)*np.exp(
                    -self.phi*rel_sed_flux_in)

        elif self.type == 'linear_decline':
            def sed_flux_fn_gen(rel_sed_flux_in):
                return 1.-rel_sed_flux_in

        elif self.type == 'parabolic':
            raise MissingKeyError(
                'Pure parabolic (where intersect at zero flux is exactly ' +
                'zero) is currently not supported, sorry. Try ' +
                'almost_parabolic instead?')

            def sed_flux_fn_gen(rel_sed_flux_in):
                return 1. - 4.*(rel_sed_flux_in-0.5)**2.

        elif self.type == 'almost_parabolic':

            def sed_flux_fn_gen(rel_sed_flux_in):
                return np.where(rel_sed_flux_in > 0.1,
                                1. - 4.*(rel_sed_flux_in-0.5)**2.,
                                2.6*rel_sed_flux_in+0.1)

        elif self.type == 'None':

            def sed_flux_fn_gen(rel_sed_flux_in):
                return 1.
        else:
            raise MissingKeyError(
                'Provided sed flux sensitivity type in input file was not ' +
                'recognised!')

        for i in range(self.pseudoimplicit_repeats):
            sed_flux_fn = sed_flux_fn_gen(rel_sed_flux)
            sed_vol_added = prefactor_for_volume*sed_flux_fn
            rel_sed_flux = rel_sed_flux_in + sed_vol_added/trans_cap_vol_out
            # print rel_sed_flux
            if rel_sed_flux >= 1.:
                rel_sed_flux = 1.
                break
            if rel_sed_flux < 0.:
                rel_sed_flux = 0.
                break
        last_sed_flux_fn = sed_flux_fn
        sed_flux_fn = sed_flux_fn_gen(rel_sed_flux)
        # this error could alternatively be used to break the loop
        error_in_sed_flux_fn = sed_flux_fn-last_sed_flux_fn
        dz = prefactor_for_dz*sed_flux_fn
        sed_flux_out = rel_sed_flux*trans_cap_vol_out
        return dz, sed_flux_out, rel_sed_flux, error_in_sed_flux_fn

    def erode(self, dt, flooded_depths=None, **kwds):
        """Erode and deposit on the channel bed for a duration of *dt*.

        Erosion occurs according to the sediment dependent rules specified
        during initialization.

        Parameters
        ----------
        dt : float (years, only!)
            Timestep for which to run the component.
        flooded_depths : array or field name (m)
            Depths of flooding at each node, zero where no lake. Note that the
            component will dynamically update this array as it fills nodes
            with sediment (...but does NOT update any other related lake
            fields).
        """
        grid = self.grid
        node_z = grid.at_node['topographic__elevation']
        node_A = grid.at_node['drainage_area']
        flow_receiver = grid.at_node['flow__receiver_node']
        s_in = grid.at_node['flow__upstream_node_order']
        node_S = grid.at_node['topographic__steepest_slope']

        if type(flooded_depths) is str:
            flooded_depths = mg.at_node[flooded_depths]
            # also need a map of initial flooded conds:
            flooded_nodes = flooded_depths > 0.
        elif type(flooded_depths) is np.ndarray:
            assert flooded_depths.size == self.grid.number_of_nodes
            flooded_nodes = flooded_depths > 0.
            # need an *updateable* record of the pit depths
        else:
            # if None, handle in loop
            flooded_nodes = None
        steepest_link = 'flow__link_to_receiver_node'
        link_length = np.empty(grid.number_of_nodes, dtype=float)
        link_length.fill(np.nan)
        draining_nodes = np.not_equal(grid.at_node[steepest_link],
                                      BAD_INDEX_VALUE)
        core_draining_nodes = np.intersect1d(np.where(draining_nodes)[0],
                                             grid.core_nodes,
                                             assume_unique=True)
        link_length[core_draining_nodes] = grid.length_of_d8[
            grid.at_node[steepest_link][core_draining_nodes]]

        if self.Qc == 'MPM':
            if self.Dchar_in is not None:
                self.Dchar = self.Dchar_in
            else:
                assert not self.set_threshold, (
                    "Something is seriously wrong with your model " +
                    "initialization.")
                assert self.override_threshold, (
                    "You need to confirm to the module you intend it to " +
                    "internally calculate a shear stress threshold, " +
                    "with set_threshold_from_Dchar in the input file.")
                # we need to adjust the thresholds for the Shields number
                # & gs dynamically:
                variable_thresh = self.shields_crit*self.g*(
                    self.sed_density-self.fluid_density)*self.Dchar
            if self.lamb_flag:
                variable_shields_crit = 0.15*node_S**0.25
                try:
                    variable_thresh = (variable_shields_crit *
                                       self.shields_prefactor_to_shear)
                except AttributeError:
                    variable_thresh = (
                        variable_shields_crit *
                        self.shields_prefactor_to_shear_noDchar*self.Dchar)

            node_Q = self.k_Q*self.runoff_rate*node_A**self._c
            shear_stress_prefactor_timesAparts = (
                self.shear_stress_prefactor*node_Q**self.point6onelessb)
            try:
                transport_capacities_thresh = (
                    self.thresh*self.Qs_thresh_prefactor*self.runoff_rate**(
                        0.66667*self._b)*node_A**self.Qs_power_onAthresh)
            except AttributeError:
                transport_capacities_thresh = (
                    variable_thresh*self.Qs_thresh_prefactor *
                    self.runoff_rate**(0.66667*self._b)*node_A **
                    self.Qs_power_onAthresh)

            transport_capacity_prefactor_withA = (
                self.Qs_prefactor*self.runoff_rate**(0.6+self._b/15.)*node_A **
                self.Qs_power_onA)

            internal_t = 0.
            break_flag = False
            dt_secs = dt*31557600.
            counter = 0
            rel_sed_flux = np.empty_like(node_Q)
            # excess_vol_overhead = 0.

            while 1:
                # ^use the break flag, to improve computational efficiency for
                # runs which are very stable
                # we assume the drainage structure is forbidden to change
                # during the whole dt
                # note slopes will be *negative* at pits
                # track how many loops we perform:
                counter += 1
                downward_slopes = node_S.clip(0.)
                # this removes the tendency to transfer material against
                # gradient, including in any lake depressions
                # we DON'T immediately zero trp capacity in the lake.
                # positive_slopes = np.greater(downward_slopes, 0.)
                slopes_tothe07 = downward_slopes**0.7
                transport_capacities_S = (transport_capacity_prefactor_withA *
                                          slopes_tothe07)
                trp_diff = (transport_capacities_S -
                            transport_capacities_thresh).clip(0.)
                transport_capacities = np.sqrt(trp_diff*trp_diff*trp_diff)
                shear_stress = (shear_stress_prefactor_timesAparts *
                                slopes_tothe07)
                shear_tothe_a = shear_stress**self._a

                dt_this_step = dt_secs-internal_t
                # ^timestep adjustment is made AFTER the dz calc
                node_vol_capacities = transport_capacities*dt_this_step

                sed_into_node = np.zeros(grid.number_of_nodes, dtype=float)
                dz = np.zeros(grid.number_of_nodes, dtype=float)
                len_s_in = s_in.size
                cell_areas = self.cell_areas
                try:
                    raise NameError
                    # ^tripped out deliberately for now; doesn't appear to
                    # accelerate much
                    weave.inline(self.routing_code, [
                        'len_s_in', 'sed_into_node', 'transport_capacities',
                        'dz', 'cell_areas', 'dt_this_step', 'flow__receiver_node'])
                except NameError:
                    for i in s_in[::-1]:  # work downstream
                        cell_area = cell_areas[i]
                        if flooded_nodes is not None:
                            flood_depth = flooded_depths[i]
                        else:
                            flood_depth = 0.
                        sed_flux_into_this_node = sed_into_node[i]
                        node_capacity = transport_capacities[i]
                        # ^we work in volume flux, not volume per se here
                        node_vol_capacity = node_vol_capacities[i]
                        if flood_depth > 0.:
                            node_vol_capacity = 0.
                            # requires special case handling - as much sed as
                            # possible is dumped here, then the remainder
                            # passed on
                        if sed_flux_into_this_node < node_vol_capacity:
                            # ^note incision is forbidden at capacity
                            # flooded nodes never enter this branch
                            # #implementing the pseudoimplicit method:
                            try:
                                thresh = variable_thresh
                            except:  # it doesn't exist
                                thresh = self.thresh
                            dz_prefactor = self._K_unit_time*dt_this_step*(
                                shear_tothe_a[i]-thresh).clip(0.)
                            vol_prefactor = dz_prefactor*cell_area
                            (dz_here, sed_flux_out, rel_sed_flux_here,
                             error_in_sed_flux) = \
                                self.get_sed_flux_function_pseudoimplicit(
                                    sed_flux_into_this_node, node_vol_capacity,
                                    vol_prefactor, dz_prefactor)
                            # note now dz_here may never create more sed than
                            # the out can transport...
                            assert sed_flux_out <= node_vol_capacity, (
                                'failed at node '+str(s_in.size-i) +
                                ' with rel sed flux '+str(
                                    sed_flux_out/node_capacity))
                            rel_sed_flux[i] = rel_sed_flux_here
                            vol_pass = sed_flux_out
                        else:
                            rel_sed_flux[i] = 1.
                            vol_dropped = (sed_flux_into_this_node -
                                           node_vol_capacity)
                            dz_here = -vol_dropped/cell_area
                            # with the pits, we aim to inhibit incision, but
                            # depo is OK. We have already zero'd any adverse
                            # grads, so sed can make it to the bottom of the
                            # pit but no further in a single step, which seems
                            # raeasonable. Pit should fill.
                            if flood_depth <= 0.:
                                vol_pass = node_vol_capacity
                            else:
                                height_excess = -dz_here - flood_depth
                                # ...above water level
                                if height_excess <= 0.:
                                    vol_pass = 0.
                                    # dz_here is already correct
                                    flooded_depths[i] += dz_here
                                else:
                                    dz_here = -flood_depth
                                    vol_pass = height_excess * cell_area
                                    # ^bit cheeky?
                                    flooded_depths[i] = 0.
                                    # note we must update flooded depths
                                    # transiently to conserve mass
                            # do we need to retain a small downhill slope?
                            # ...don't think so. Will resolve itself on next
                            # timestep.

                        dz[i] -= dz_here
                        sed_into_node[flow_receiver[i]] += vol_pass

                break_flag = True

                node_z[grid.core_nodes] += dz[grid.core_nodes]

                if break_flag:
                    break
                # do we need to reroute the flow/recalc the slopes here?
                # -> NO, slope is such a minor component of Diff we'll be OK
                # BUT could be important not for the stability, but for the
                # actual calc. So YES.
                node_S = np.zeros_like(node_S)
                node_S[core_draining_nodes] = (node_z-node_z[flow_receiver])[
                    core_draining_nodes]/link_length[core_draining_nodes]
                internal_t += dt_this_step  # still in seconds, remember

        elif self.Qc == 'power_law':
            transport_capacity_prefactor_withA = self._Kt * node_A**self._mt
            erosion_prefactor_withA = self._K_unit_time * node_A**self._m
            # ^doesn't include S**n*f(Qc/Qc)
            internal_t = 0.
            break_flag = False
            dt_secs = dt*31557600.
            counter = 0
            rel_sed_flux = np.empty_like(node_A)
            while 1:
                counter += 1
                # print counter
                downward_slopes = node_S.clip(0.)
                # positive_slopes = np.greater(downward_slopes, 0.)
                slopes_tothen = downward_slopes**self._n
                slopes_tothent = downward_slopes**self._nt
                transport_capacities = (transport_capacity_prefactor_withA *
                                        slopes_tothent)
                erosion_prefactor_withS = (
                    erosion_prefactor_withA * slopes_tothen)  # no time, no fqs
                # shear_tothe_a = shear_stress**self._a

                dt_this_step = dt_secs-internal_t
                # ^timestep adjustment is made AFTER the dz calc
                node_vol_capacities = transport_capacities*dt_this_step

                sed_into_node = np.zeros(grid.number_of_nodes, dtype=float)
                dz = np.zeros(grid.number_of_nodes, dtype=float)
                cell_areas = self.cell_areas
                for i in s_in[::-1]:  # work downstream
                    cell_area = cell_areas[i]
                    if flooded_nodes is not None:
                        flood_depth = flooded_depths[i]
                    else:
                        flood_depth = 0.
                    sed_flux_into_this_node = sed_into_node[i]
                    node_capacity = transport_capacities[i]
                    # ^we work in volume flux, not volume per se here
                    node_vol_capacity = node_vol_capacities[i]
                    if flood_depth > 0.:
                        node_vol_capacity = 0.
                    if sed_flux_into_this_node < node_vol_capacity:
                        # ^note incision is forbidden at capacity
                        dz_prefactor = dt_this_step*erosion_prefactor_withS[i]
                        vol_prefactor = dz_prefactor*cell_area
                        (dz_here, sed_flux_out, rel_sed_flux_here,
                         error_in_sed_flux) = \
                            self.get_sed_flux_function_pseudoimplicit(
                                sed_flux_into_this_node, node_vol_capacity,
                                vol_prefactor, dz_prefactor)
                        # note now dz_here may never create more sed than the
                        # out can transport...
                        assert sed_flux_out <= node_vol_capacity, (
                            'failed at node '+str(s_in.size-i) +
                            ' with rel sed flux '+str(
                                sed_flux_out/node_capacity))
                        rel_sed_flux[i] = rel_sed_flux_here
                        vol_pass = sed_flux_out
                    else:
                        rel_sed_flux[i] = 1.
                        vol_dropped = (sed_flux_into_this_node -
                                       node_vol_capacity)
                        dz_here = -vol_dropped/cell_area
                        try:
                            isflooded = flooded_nodes[i]
                        except TypeError:  # was None
                            isflooded = False
                        if flood_depth <= 0. and not isflooded:
                            vol_pass = node_vol_capacity
                            # we want flooded nodes which have already been
                            # filled to enter the else statement
                        else:
                            height_excess = -dz_here - flood_depth
                            # ...above water level
                            if height_excess <= 0.:
                                vol_pass = 0.
                                # dz_here is already correct
                                flooded_depths[i] += dz_here
                            else:
                                dz_here = -flood_depth
                                vol_pass = height_excess * cell_area
                                # ^bit cheeky?
                                flooded_depths[i] = 0.

                    dz[i] -= dz_here
                    sed_into_node[flow_receiver[i]] += vol_pass
                break_flag = True

                node_z[grid.core_nodes] += dz[grid.core_nodes]

                if break_flag:
                    break
                # do we need to reroute the flow/recalc the slopes here?
                # -> NO, slope is such a minor component of Diff we'll be OK
                # BUT could be important not for the stability, but for the
                # actual calc. So YES.
                node_S = np.zeros_like(node_S)
                # print link_length[core_draining_nodes]
                node_S[core_draining_nodes] = (node_z-node_z[flow_receiver])[
                    core_draining_nodes]/link_length[core_draining_nodes]
                internal_t += dt_this_step  # still in seconds, remember

        active_nodes = grid.core_nodes

        if self.return_ch_props:
            # add the channel property field entries,
            # 'channel__width', 'channel__depth', and 'channel__discharge'
            W = self.k_w*node_Q**self._b
            H = shear_stress/self.rho_g/node_S  # ...sneaky!
            grid.at_node['channel__width'][:] = W
            grid.at_node['channel__depth'][:] = H
            grid.at_node['channel__discharge'][:] = node_Q
            grid.at_node['channel__bed_shear_stress'][:] = shear_stress

        grid.at_node['channel_sediment__volumetric_transport_capacity'][
            :] = transport_capacities
        grid.at_node['channel_sediment__volumetric_flux'][:] = sed_into_node
        grid.at_node['channel_sediment__relative_flux'][:] = rel_sed_flux
        # elevs set automatically to the name used in the function call.
        self.iterations_in_dt = counter

        return grid, grid.at_node['topographic__elevation']

    def run_one_step(self, dt, flooded_depths=None, **kwds):
        """Run the component across one timestep increment, dt.

        Erosion occurs according to the sediment dependent rules specified
        during initialization. Method is fully equivalent to the :func:`erode`
        method.

        Parameters
        ----------
        dt : float (years, only!)
            Timestep for which to run the component.
        flooded_depths : array or field name (m)
            Depths of flooding at each node, zero where no lake. Note that the
            component will dynamically update this array as it fills nodes
            with sediment (...but does NOT update any other related lake
            fields).
        """
        self.erode(dt=dt, flooded_depths=flooded_depths, **kwds)

    @property
    @make_return_array_immutable
    def characteristic_grainsize(self):
        """Return the characteristic grain size used by the component.

        Particularly useful if the set_Dchar_from_threshold flag was True
        at initialization.

        Returns
        -------
        Dchar : float or array

        Examples
        --------
        >>> from landlab import RasterModelGrid
        >>> mg1 = RasterModelGrid((3,4), 1.)
        >>> thresh_shields = np.arange(1, mg1.number_of_nodes+1, dtype=float)
        >>> thresh_shields /= 100.
        >>> sde1 = SedDepEroder(mg1, threshold_shear_stress=100., Qc='MPM',
        ...                     Dchar=None, set_threshold_from_Dchar=False,
        ...                     set_Dchar_from_threshold=True,
        ...                     threshold_Shields=thresh_shields)
        >>> sde1.characteristic_grainsize
        array([ 0.59962823,  0.29981412,  0.19987608,  0.14990706,  0.11992565,
                0.09993804,  0.08566118,  0.07495353,  0.06662536,  0.05996282,
                0.05451166,  0.04996902])

        >>> mg2 = RasterModelGrid((3,4), 1.)
        >>> sde2 = SedDepEroder(mg2, threshold_shear_stress=100., Qc='MPM',
        ...                     Dchar=None, set_threshold_from_Dchar=False,
        ...                     set_Dchar_from_threshold=True,
        ...                     threshold_Shields=None,
        ...                     slope_sensitive_threshold=True)
        >>> S = mg2.add_ones('node', 'topographic__steepest_slope')
        >>> S *= 0.05  # thresh = 100 Pa @ 5pc slope
        >>> sde2.characteristic_grainsize  # doctest: +NORMALIZE_WHITESPACE
        array([ 0.08453729,  0.08453729,  0.08453729,  0.08453729,
                0.08453729,  0.08453729,  0.08453729,  0.08453729,
                0.08453729,  0.08453729,  0.08453729,  0.08453729])
        """
        # Dchar is None means self.lamb_flag, Dchar is spatially variable,
        # and not calculated until the main loop
        assert self.Qc == 'MPM', ("Characteristic grainsize is only " +
                                  "calculated if Qc == 'MPM'")
        if self.Dchar_in is not None:
            return self.Dchar_in
        else:
            taustarcrit = 0.15 * self.grid.at_node[
                'topographic__steepest_slope']**0.25
            Dchar = self.thresh/(self.g*(self.sed_density -
                                         self.fluid_density)*taustarcrit)
            return Dchar



import numpy as np
from six.moves import range

from ...utils.jaggedarray import JaggedArray


class CellGrid(object):
    """
    Parameters
    ----------
    vertices : array-like
        Vertex IDs at each grid cell
    vertices_per_cell : array-like
        Number of vertices per grid cell

    Returns
    -------
    CellGrid :
        A newly-created CellGrid

    Examples
    --------
    Create a grid of two cells where the first cell has four vertices and
    the second has three.

    >>> from landlab.grid.unstructured.cells import CellGrid
    >>> cgrid = CellGrid([0, 1, 3, 2, 1, 4, 3], [4, 3])
    >>> cgrid.number_of_cells
    2
    >>> cgrid.number_of_vertices_at_cell(0)
    4
    >>> cgrid.number_of_vertices_at_cell(1)
    3
    >>> cgrid.vertices_at_cell(0)
    array([0, 1, 3, 2])
    >>> cgrid.vertices_at_cell(1)
    array([1, 4, 3])

    Associate nodes with each cell.

    >>> cgrid = CellGrid([0, 1, 2, 3, 1, 3, 4], [4, 3], node_at_cell=[10, 11])
    >>> cgrid.node_at_cell
    array([10, 11])
    >>> cgrid.cell_at_node[10]
    0
    >>> cgrid.cell_at_node[11]
    1
    """

    def __init__(self, vertices, vertices_per_cell, node_at_cell=None):
        """
        Parameters
        ----------
        vertices : array-like
            Vertex IDs at each grid cell
        vertices_per_cell : array-like
            Number of vertices per grid cell

        Returns
        -------
        CellGrid :
            A newly-created CellGrid

        Examples
        --------
        Create a grid of two cells where the first cell has four vertices and
        the second has three.

        >>> from landlab.grid.unstructured.cells import CellGrid
        >>> cgrid = CellGrid([0, 1, 3, 2, 1, 4, 3], [4, 3])
        >>> cgrid.number_of_cells
        2
        >>> cgrid.number_of_vertices_at_cell(0)
        4
        >>> cgrid.number_of_vertices_at_cell(1)
        3
        >>> cgrid.vertices_at_cell(0)
        array([0, 1, 3, 2])
        >>> cgrid.vertices_at_cell(1)
        array([1, 4, 3])

        Associate nodes with each cell.

        >>> cgrid = CellGrid([0, 1, 2, 3, 1, 3, 4], [4, 3], node_at_cell=[10, 11])
        >>> cgrid.node_at_cell
        array([10, 11])
        >>> cgrid.cell_at_node[10]
        0
        >>> cgrid.cell_at_node[11]
        1
        """
        self._vertices_at_cell = JaggedArray(vertices, vertices_per_cell)
        self._number_of_cells = len(vertices_per_cell)

        if node_at_cell:
            self._node_at_cell = np.array(node_at_cell)
            self._cell_at_node = np.ma.masked_all(
                max(node_at_cell) + 1, dtype=int)
            self._cell_at_node[self._node_at_cell] = range(len(node_at_cell))
            #self._cell_id_map = dict(zip(node_at_cell, range(len(node_at_cell))))

    @property
    def number_of_cells(self):
        return self._number_of_cells

    @property
    def node_at_cell(self):
        return self._node_at_cell

    @property
    def cell_at_node(self):
        return self._cell_at_node

    def number_of_vertices_at_cell(self, cell):
        return self._vertices_at_cell.length_of_row(cell)

    def vertices_at_cell(self, cell):
        return self._vertices_at_cell.row(cell)

    def iter(self):
        """Iterate over the cells of the grid.

        Returns
        -------
        ndarray :
            Nodes entering and leaving each node
        """
        for cell in range(self.number_of_cells):
            yield self.vertices_at_cell(cell)

    @property
    def cell_id(self):
        try:
            return self._cell_ids
        except AttributeError:
            return np.arange(self.number_of_cells)

    def vertices_at_cell_id(self, cell_id):
        try:
            return self.vertices_at_cell(self._cell_id_map[cell_id])
        except AttributeError:
            return self.vertices_at_cell(cell_id)



import numpy as np
cimport numpy as np
cimport cython
from libc.stdlib cimport malloc, free


DTYPE = np.int
ctypedef np.int_t DTYPE_t


def _is_finite_region(np.ndarray[DTYPE_t, ndim=1] vertices_at_region,
                      np.ndarray[DTYPE_t, ndim=1] vertices_per_region,
                      np.ndarray[DTYPE_t, ndim=1] is_finite_region,
                      DTYPE_t min_patch_size):
    """Test if each region if finite.

    Parameters
    ----------
    vertices_at_region : ndarray of int
        Indices of the Voronoi vertices forming each Voronoi region.
    vertices_per_region : ndarray of int, shape `(n_regions, )`
        Number of vertices that form each voronoi region.
    is_finite_region : ndarray of int, shape `(n_regions, )`
        Output buffer. 1 if a region is finite, otherwise 0.
    """
    cdef int i
    cdef int region
    cdef int vertex
    cdef int n_vertices
    cdef int offset = 0
    cdef int n_regions = len(vertices_per_region)

    for region in range(n_regions):
        n_vertices = vertices_per_region[region]
        if n_vertices == 0 or n_vertices < min_patch_size:
            is_finite_region[region] = 0
        else:
            is_finite_region[region] = 1
            for i in range(n_vertices):
                vertex = vertices_at_region[offset + i]
                if vertex == -1:
                    is_finite_region[region] = 0
                    break
        offset += n_vertices


def _get_neighbor_regions(np.ndarray[DTYPE_t, ndim=2] ridge_points,
                          np.ndarray[DTYPE_t, ndim=1] point_region,
                          np.ndarray[DTYPE_t, ndim=1] is_finite_region,
                          np.ndarray[DTYPE_t, ndim=2] regions_at_ridge):
    """Get voronoi regions on either side of ridges.

    Parameters
    ----------
    ridge_points : ndarray of int, shape `(n_points, 2)`
        Indices of the points between which each Voronoi ridge lies (as
        provided by the `scipy.spatial.Voronoi` class).
    point_region : ndarray of int, shape `(n_points, )`
        Index of the Voronoi region for each input point (as provided by the
        `scipy.spatial.Voronoi` class).
    is_finite_region : ndarray of int, shape `(n_regions, )`
        1 if a region is finite, otherwise 0.
    regions_at_ridge : ndarray of int, shape `(n_ridges, 2)`
        Indices of the voronoi regions on either side of each ridge.
    """
    cdef int i
    cdef int ridge
    cdef int region
    cdef int n_ridges = len(ridge_points)

    for ridge in range(n_ridges):
        for i in range(2):
            region = point_region[ridge_points[ridge, i]]
            if is_finite_region[region]:
                regions_at_ridge[ridge, i] = region
            else:
                regions_at_ridge[ridge, i] = -1


@cython.boundscheck(False)
def _get_cell_at_region(np.ndarray[DTYPE_t, ndim=2] regions_at_ridge,
                        np.ndarray[DTYPE_t, ndim=2] ridges_at_cell,
                        np.ndarray[DTYPE_t, ndim=1] cell_at_region):
    """Get cell corresponding to each voronoi region.

    Parameters
    ----------
    regions_at_ridge : ndarray of int, shape `(n_regions, 2)`
        Indices of the voronoi regions on either side of each ridge.
    ridges_at_cell : ndarray of int, shape `(n_cells, 2)`
        Output buffer. The voronoi regions on either side of each cell.
    cell_at_region : ndarray of int, shape `(n_regions, )`
        Output buffer. ID of the cell corresponding to each voronoi region.
        Regions without cells are set to -1.
    """
    cdef int i
    cdef int ridge
    cdef int cell
    cdef int region
    cdef int n_cells = 0
    cdef int n_ridges = len(regions_at_ridge)
    cdef int *ridges_per_cell = <int *>malloc(n_ridges * sizeof(int))

    if not ridges_per_cell:
        raise MemoryError('unable to allocate {bytes} bytes'.format(
            bytes=n_ridges * sizeof(int)))

    try:
        for ridge in range(n_ridges):
            ridges_per_cell[ridge] = 0

        for region in range(cell_at_region.shape[0]):
            cell_at_region[region] = -1

        for cell in range(ridges_at_cell.shape[0]):
            for ridge in range(ridges_at_cell.shape[1]):
                ridges_at_cell[cell, ridge] = -1

        for ridge in range(n_ridges):
            for i in range(2):
                region = regions_at_ridge[ridge, i]
                if region >= 0:
                    if cell_at_region[region] == -1:
                        cell_at_region[region] = n_cells
                        n_cells += 1
                    cell = cell_at_region[region]

                    ridges_at_cell[cell, ridges_per_cell[cell]] = ridge
                    ridges_per_cell[cell] += 1

    finally:
        free(ridges_per_cell)

    return n_cells


@cython.boundscheck(False)
def _get_faces_at_cell(np.ndarray[DTYPE_t, ndim=2] ridges_at_cell,
                       np.ndarray[DTYPE_t, ndim=2] faces_at_cell,
                       np.ndarray[DTYPE_t, ndim=1] face_at_ridge):
    """Get faces that define each cell.

    Parameters
    ----------
    ridges_at_cell : ndarray of int, shape `(n_cells, max_ridges)`
        Indices of the voronoi ridges that define each cell. Cells with less
        than `max_ridges` ridges are padded with -1.
    faces_at_cell : ndarray of int, shape `(n_cells, max_ridges)`
        Output buffer. IDs of faces that define each cell. Cells with less
        than `max_ridges` faces are padded with -1.
    face_at_ridge : ndarray of int, shape `(n_ridges, )`
        Output buffer. ID of face corresponding to each voronoi ridge. Ridges
        without a face are set to -1.
    """
    cdef int i
    cdef int face
    cdef int cell
    cdef int ridge
    cdef int n_faces
    cdef int n_ridges = len(face_at_ridge)
    cdef int n_cells = ridges_at_cell.shape[0]
    cdef int max_faces_per_cell = ridges_at_cell.shape[1]

    for i in range(n_ridges):
        face_at_ridge[i] = -1

    face = 0
    n_faces = 0
    for cell in range(n_cells):
        for i in range(max_faces_per_cell):
            ridge = ridges_at_cell[cell, i]
            if ridge >= 0:
                face = face_at_ridge[ridge]

                if face == -1:
                  face_at_ridge[ridge] = n_faces
                  n_faces += 1
                face = face_at_ridge[ridge]

                faces_at_cell[cell, i] = face
                face_at_ridge[ridge] = face
            else:
                faces_at_cell[cell, i] = -1


@cython.boundscheck(False)
def _get_corners_at_face(np.ndarray[DTYPE_t, ndim=1] face_at_ridge,
                         np.ndarray[DTYPE_t, ndim=2] vertices_at_ridge,
                         np.ndarray[DTYPE_t, ndim=1] corner_at_vertex,
                         np.ndarray[DTYPE_t, ndim=2] corners_at_face):
    """Get corners for each face.

    Parameters
    ----------
    face_at_ridge : ndarray of int, shape `(n_ridges, )`
        ID of the face corresponding to each ridge (-1 for ridges without
        a face). A ridge will not have a face if it is appears only in infinite
        regions.
    vertices_at_ridge : ndarray of int, shape `(n_ridges, 2)`
        Indices of the Voronoi vertices forming each Voronoi ridge (as provided
        by the `ridge_vertices` attribute of `scipy.spatial.Voronoi.class`.
    corner_at_vertex : ndarray of int, shape `(n_vertices, )`
        Output buffer. ID of corner corresponding to each voronoi vertex.
    corners_at_face : ndarray of int, shape `(n_vertices, 2)`
        Output buffer. IDs of two corners that define each face.
    """
    cdef int i
    cdef int ridge
    cdef int corner
    cdef int vertex
    cdef int face
    cdef int n_ridges = len(face_at_ridge)
    cdef int n_vertices = len(corner_at_vertex)
    cdef int n_faces = len(corners_at_face)
    cdef int n_corners = 0

    for vertex in range(n_vertices):
      corner_at_vertex[vertex] = -1

    for face in range(n_faces):
      for i in range(2):
        corners_at_face[face, i] = -1

    for ridge in range(n_ridges):
        for i in range(2):
            vertex = vertices_at_ridge[ridge, i]
            if vertex >= 0 and face_at_ridge[ridge] >= 0:
              if corner_at_vertex[vertex] == -1:
                  corner_at_vertex[vertex] = n_corners
                  n_corners += 1

              corners_at_face[face_at_ridge[ridge], i] = corner_at_vertex[vertex]


@cython.boundscheck(False)
def _get_xy_at_corners(np.ndarray[np.double_t, ndim=2] vertices,
                       np.ndarray[DTYPE_t, ndim=1] corner_at_vertex,
                       np.ndarray[np.double_t, ndim=2] xy_at_corner):
    """Get x and y coordinates for each corner.

    Parameters
    ----------
    vertices : ndarray of double, shape `(n_vertices, 2)`
        Coordinates of the Voronoi vertices (as provided by the
        `scipy.spacial.Voronoi` class.
    corner_at_vertex : ndarray of int, shape `(n_vertices, )`
        ID of the corner corresponding to each voronoi vertex.
    xy_at_corner : ndarray of double, shape `(n_vertices, 2)`
        Output buffer. Coordinates of the corners (first column is *x*,
        second column is *y*).
    """
    cdef int vertex
    cdef int corner
    cdef int n_vertices = len(vertices)

    for vertex in range(n_vertices):
        corner = corner_at_vertex[vertex]

        if corner >= 0:
          xy_at_corner[corner, 0] = vertices[vertex, 0]
          xy_at_corner[corner, 1] = vertices[vertex, 1]


@cython.boundscheck(False)
def _get_node_at_cell(np.ndarray[DTYPE_t, ndim=2] ridge_points,
                      np.ndarray[DTYPE_t, ndim=1] point_region,
                      np.ndarray[DTYPE_t, ndim=1] cell_at_region,
                      np.ndarray[DTYPE_t, ndim=1] node_at_cell):
    """Get node-to-cell connectivity.

    Parameters
    ----------
    ridge_points : ndarray of int, shape `(n_ridges, 2)`
        Indices of the points between which each Voronoi ridge lies (as
        provided by the `scipy.spatial.Voronoi` class).
    point_region : ndarray of int, shape `(n_points, )`
        Index of the Voronoi region for each input point (as provided by the
        `scipy.spatial.Voronoi` class).
    cell_at_region : ndarray of int, shape `(n_regions, )`
        ID of the cell corresponding to each voronoi region.
    node_at_cell : ndarray of int, shape `(n_cells, )`
        Output buffer. The node corresponding to each cell.
    """
    cdef int point
    cdef int region
    cdef int cell
    cdef int n_regions = len(point_region)

    for point in range(n_regions):
        region = point_region[point]
        cell = cell_at_region[region]
        if cell >= 0:
            node_at_cell[cell] = point



#! /usr/bin/env python

from setuptools import setup, find_packages, Extension
from setuptools.command.install import install
from setuptools.command.develop import develop
from distutils.extension import Extension
import pkg_resources

import versioneer


numpy_incl = pkg_resources.resource_filename('numpy', 'core/include')


ext_modules = [
    Extension('landlab.ca.cfuncs',
              ['landlab/ca/cfuncs.pyx']),
    Extension('landlab.grid.cfuncs',
              ['landlab/grid/cfuncs.pyx']),
    Extension('landlab.components.flexure.cfuncs',
              ['landlab/components/flexure/cfuncs.pyx']),
    Extension('landlab.components.flexure.ext.flexure1d',
              ['landlab/components/flexure/ext/flexure1d.pyx']),
    Extension('landlab.components.flow_accum.cfuncs',
              ['landlab/components/flow_accum/cfuncs.pyx']),
    Extension('landlab.components.flow_director.cfuncs',
              ['landlab/components/flow_director/cfuncs.pyx']),
    Extension('landlab.components.stream_power.cfuncs',
              ['landlab/components/stream_power/cfuncs.pyx']),
    Extension('landlab.components.space.cfuncs',
              ['landlab/components/space/cfuncs.pyx']),
    Extension('landlab.components.drainage_density.cfuncs',
              ['landlab/components/drainage_density/cfuncs.pyx']),
    Extension('landlab.components.erosion_deposition.cfuncs',
              ['landlab/components/erosion_deposition/cfuncs.pyx']),
    Extension('landlab.utils.ext.jaggedarray',
              ['landlab/utils/ext/jaggedarray.pyx']),
    Extension('landlab.graph.structured_quad.ext.at_node',
              ['landlab/graph/structured_quad/ext/at_node.pyx']),
    Extension('landlab.graph.structured_quad.ext.at_link',
              ['landlab/graph/structured_quad/ext/at_link.pyx']),
    Extension('landlab.graph.structured_quad.ext.at_patch',
              ['landlab/graph/structured_quad/ext/at_patch.pyx']),
    Extension('landlab.graph.structured_quad.ext.at_cell',
              ['landlab/graph/structured_quad/ext/at_cell.pyx']),
    Extension('landlab.graph.structured_quad.ext.at_face',
              ['landlab/graph/structured_quad/ext/at_face.pyx']),
    Extension('landlab.graph.hex.ext.hex',
              ['landlab/graph/hex/ext/hex.pyx']),
    Extension('landlab.graph.sort.ext.remap_element',
              ['landlab/graph/sort/ext/remap_element.pyx']),
    Extension('landlab.graph.sort.ext.argsort',
              ['landlab/graph/sort/ext/argsort.pyx']),
    Extension('landlab.graph.sort.ext.spoke_sort',
              ['landlab/graph/sort/ext/spoke_sort.pyx']),
    Extension('landlab.graph.voronoi.ext.voronoi',
              ['landlab/graph/voronoi/ext/voronoi.pyx']),
    Extension('landlab.graph.voronoi.ext.delaunay',
              ['landlab/graph/voronoi/ext/delaunay.pyx']),
    Extension('landlab.graph.object.ext.at_node',
              ['landlab/graph/object/ext/at_node.pyx']),
    Extension('landlab.graph.object.ext.at_patch',
              ['landlab/graph/object/ext/at_patch.pyx']),
    Extension('landlab.graph.quantity.ext.of_link',
              ['landlab/graph/quantity/ext/of_link.pyx']),
    Extension('landlab.graph.quantity.ext.of_patch',
              ['landlab/graph/quantity/ext/of_patch.pyx']),
    Extension('landlab.graph.matrix.ext.matrix',
              ['landlab/graph/matrix/ext/matrix.pyx']),
    Extension('landlab.grid.structured_quad.cfuncs',
              ['landlab/grid/structured_quad/cfuncs.pyx']),
    Extension('landlab.grid.structured_quad.c_faces',
              ['landlab/grid/structured_quad/c_faces.pyx']),
    Extension('landlab.layers.ext.eventlayers',
              ['landlab/layers/ext/eventlayers.pyx']),
]


def register(**kwds):
    import httplib, urllib

    data = urllib.urlencode(kwds)
    header = {"Content-type": "application/x-www-form-urlencoded",
              "Accept": "text/plain"}
    conn = httplib.HTTPConnection('csdms.colorado.edu')
    conn.request('POST', '/register/', data, header)


def register_landlab():
    try:
        from sys import argv
        import platform
        data = {
            'name': 'landlab',
            'version': __version__,
            'platform': platform.platform(),
            'desc': ';'.join(argv),
        }
        register(**data)
    except Exception:
        pass


class install_and_register(install):
    def run(self):
        install.run(self)
        register_landlab()


class develop_and_register(develop):
    def run(self):
        develop.run(self)
        register_landlab()


import os

#cython_pathspec = os.path.join('landlab', 'components','**','*.pyx')
#ext_modules = cythonize(cython_pathspec)


setup(name='landlab',
      version=versioneer.get_version(),
      author='Eric Hutton',
      author_email='eric.hutton@colorado.edu',
      url='https://github.com/landlab',
      description='Plugin-based component modeling tool.',
      long_description=open('README.rst').read(),
      setup_requires=['cython'],
      classifiers=[
          'Intended Audience :: Science/Research',
          'License :: OSI Approved :: MIT License',
          'Operating System :: OS Independent',
          'Programming Language :: Cython',
          'Programming Language :: Python :: 2.6',
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: Implementation :: CPython',
          'Topic :: Scientific/Engineering :: Physics'
      ],
      packages=find_packages(),
      package_data={'': ['tests/*txt', 'data/*asc', 'data/*nc',
                         'preciptest.in']},
      cmdclass=versioneer.get_cmdclass({
          'install': install_and_register,
          'develop': develop_and_register,
      }),
      entry_points={
          'console_scripts': [
              'landlab=landlab.cmd.landlab:main',
          ]
      },
      include_dirs = [numpy_incl, ],
      ext_modules = ext_modules,
     )



import numpy as np
cimport numpy as np
cimport cython

DTYPE = np.int
ctypedef np.int_t DTYPE_t


@cython.boundscheck(False)
def fill_node_at_cell(shape, np.ndarray[DTYPE_t, ndim=1] node_at_cell):
    """Get node contained in a cell.

    Parameters
    ----------
    shape : tuple of int
        Shape of the grid as `(n_rows, n_cols)`.
    node_at_cell : ndarray of int
        Buffer into which to place node identifiers.
    """
    cdef int cell
    cdef int cell_rows = shape[0] - 2
    cdef int cell_cols = shape[1] - 2
    cdef int node_cols = shape[1]
    cdef int row_offset
    cdef int row
    cdef int col

    cell = 0
    row_offset = shape[1] + 1
    for row in range(cell_rows):
        for col in range(cell_cols):
            node_at_cell[cell] = row_offset + col
            cell += 1
        row_offset += node_cols



import numpy as np
cimport numpy as np
cimport cython

DTYPE = np.int
ctypedef np.int_t DTYPE_t


@cython.boundscheck(False)
def fill_perimeter_nodes(shape, np.ndarray[DTYPE_t, ndim=1] perimeter_nodes):
    cdef int n_rows = shape[0]
    cdef int n_cols = shape[1]
    cdef int n_nodes = n_rows * n_cols
    cdef int i
    cdef int node

    # Right edge
    i = 0
    for node in range(n_cols - 1, n_nodes - 1, n_cols):
        perimeter_nodes[i] = node
        i += 1

    # Top edge
    for node in range(n_nodes - 1, n_nodes - n_cols, - 1):
        perimeter_nodes[i] = node
        i += 1

    # Left edge
    for node in range((n_rows - 1) * n_cols, 0, - n_cols):
        perimeter_nodes[i] = node
        i += 1

    # Bottom edge
    for node in range(0, n_cols - 1):
        perimeter_nodes[i] = node
        i += 1


@cython.boundscheck(False)
def fill_patches_at_node(shape, np.ndarray[DTYPE_t, ndim=2] patches_at_face):
    cdef int patch
    cdef int node
    cdef int row
    cdef int col
    cdef int n_rows = shape[0]
    cdef int n_cols = shape[1]
    cdef int patches_per_row = n_cols - 1

    # Bottom row
    patch = 0
    for node in range(1, n_cols - 1):
        patches_at_face[node, 0] = patch + 1
        patches_at_face[node, 1] = patch
        patches_at_face[node, 2] = - 1
        patches_at_face[node, 3] = - 1

        patch += 1

    # Interior nodes
    for row in range(1, n_rows - 1):
        patch = patches_per_row * row
        node = row * n_cols + 1
        for col in range(1, n_cols - 1):
            patches_at_face[node, 0] = patch + 1
            patches_at_face[node, 1] = patch
            patches_at_face[node, 2] = patch - patches_per_row
            patches_at_face[node, 3] = patch - patches_per_row + 1

            node += 1
            patch += 1

    # Top row
    patch = patches_per_row * (n_rows - 2)
    for node in range(n_cols * (n_rows - 1) + 1, n_rows * n_cols - 1):
        patches_at_face[node, 0] = - 1
        patches_at_face[node, 1] = - 1
        patches_at_face[node, 2] = patch
        patches_at_face[node, 3] = patch + 1

        patch += 1

    # Left column
    patch = patches_per_row
    for node in range(n_cols, (n_rows - 1) * n_cols, n_cols):
        patches_at_face[node, 0] = patch
        patches_at_face[node, 1] = - 1
        patches_at_face[node, 2] = - 1
        patches_at_face[node, 3] = patch - patches_per_row

        patch += patches_per_row

    # Right columns
    patch = patches_per_row - 1
    for node in range(2 * n_cols - 1, (n_rows - 1) * n_cols, n_cols):
        patches_at_face[node, 0] = - 1
        patches_at_face[node, 1] = patch + patches_per_row
        patches_at_face[node, 2] = patch
        patches_at_face[node, 3] = - 1

        patch += patches_per_row

    # Corners
    node = 0
    patches_at_face[node, 0] = 0
    patches_at_face[node, 1] = - 1
    patches_at_face[node, 2] = - 1
    patches_at_face[node, 3] = - 1

    node = n_cols - 1
    patches_at_face[node, 0] = - 1
    patches_at_face[node, 1] = patches_per_row - 1
    patches_at_face[node, 2] = - 1
    patches_at_face[node, 3] = - 1

    node = (n_rows - 1) * n_cols
    patches_at_face[node, 0] = - 1
    patches_at_face[node, 1] = - 1
    patches_at_face[node, 2] = - 1
    patches_at_face[node, 3] = patches_per_row * (n_rows - 2)

    node = n_rows * n_cols - 1
    patches_at_face[node, 0] = - 1
    patches_at_face[node, 1] = - 1
    patches_at_face[node, 2] = patches_per_row * (n_rows - 1) - 1
    patches_at_face[node, 3] = - 1


@cython.boundscheck(False)
def fill_links_at_node(shape, np.ndarray[DTYPE_t, ndim=2] links_at_node):
    cdef int n_rows = shape[0]
    cdef int n_cols = shape[1]
    cdef int n_nodes = n_rows * n_cols
    cdef int links_per_row = 2 * n_cols - 1
    cdef int patches_per_row = n_cols - 1

    # Bottom nodes
    link = 0
    for node in range(1, n_cols - 1):
        links_at_node[node, 0] = link + 1
        links_at_node[node, 1] = link + n_cols
        links_at_node[node, 2] = link
        links_at_node[node, 3] = - 1

        link += 1

    # Interior nodes
    for row in range(1, n_rows - 1):
        link = row * links_per_row
        node = row * n_cols + 1
        for col in range(1, n_cols - 1):
            links_at_node[node, 0] = link + 1
            links_at_node[node, 1] = link + n_cols
            links_at_node[node, 2] = link
            links_at_node[node, 3] = link - (n_cols - 1)

            node += 1
            link += 1

    # Top nodes
    link = (n_rows - 1) * links_per_row
    for node in range(n_nodes - n_cols + 1, n_nodes - 1):
        links_at_node[node, 0] = link + 1
        links_at_node[node, 1] = - 1
        links_at_node[node, 2] = link
        links_at_node[node, 3] = link - (n_cols - 1)

        link += 1

    # Left nodes
    link = links_per_row - 1
    for node in range(n_cols, n_nodes, n_cols):
        links_at_node[node, 0] = link + 1
        links_at_node[node, 1] = link + n_cols
        links_at_node[node, 2] = - 1
        links_at_node[node, 3] = link - (n_cols - 1)

        link += links_per_row

    # Right nodes
    link = links_per_row + (n_cols - 2)
    for node in range(2 * n_cols - 1, n_nodes, n_cols):
        links_at_node[node, 0] = - 1
        links_at_node[node, 1] = link + n_cols
        links_at_node[node, 2] = link
        links_at_node[node, 3] = link - (n_cols - 1)

        link += links_per_row

    node = 0
    links_at_node[node, 0] = 0
    links_at_node[node, 1] = n_cols - 1
    links_at_node[node, 2] = - 1
    links_at_node[node, 3] = - 1

    node = n_cols - 1
    links_at_node[node, 0] = - 1
    links_at_node[node, 1] = 2 * n_cols - 2
    links_at_node[node, 2] = n_cols - 2
    links_at_node[node, 3] = - 1

    node = n_nodes - n_cols
    links_at_node[node, 0] = (n_rows - 1) * links_per_row
    links_at_node[node, 1] = - 1
    links_at_node[node, 2] = - 1
    links_at_node[node, 3] = (n_rows - 1) * links_per_row - n_cols

    node = n_nodes - 1
    links_at_node[node, 0] = - 1
    links_at_node[node, 1] = - 1
    links_at_node[node, 2] = n_rows * links_per_row - n_cols - 1
    links_at_node[node, 3] = (n_rows - 1) * links_per_row - 1


@cython.boundscheck(False)
def fill_link_dirs_at_node(shape,
                           np.ndarray[DTYPE_t, ndim=2] link_dirs_at_node):
    cdef int n_rows = shape[0]
    cdef int n_cols = shape[1]
    cdef int n_nodes = n_rows * n_cols
    cdef int links_per_row = 2 * n_cols - 1
    cdef int patches_per_row = n_cols - 1

    # Bottom nodes
    for node in range(1, n_cols - 1):
        link_dirs_at_node[node, 0] = - 1
        link_dirs_at_node[node, 1] = - 1
        link_dirs_at_node[node, 2] = 1
        link_dirs_at_node[node, 3] = 0

    # Interior nodes
    for row in range(1, n_rows - 1):
        node = row * n_cols + 1
        for col in range(1, n_cols - 1):
            link_dirs_at_node[node, 0] = - 1
            link_dirs_at_node[node, 1] = - 1
            link_dirs_at_node[node, 2] = 1
            link_dirs_at_node[node, 3] = 1

            node += 1

    # Top nodes
    for node in range(n_nodes - n_cols + 1, n_nodes - 1):
        link_dirs_at_node[node, 0] = -1
        link_dirs_at_node[node, 1] = 0
        link_dirs_at_node[node, 2] = 1
        link_dirs_at_node[node, 3] = 1

    # Left nodes
    for node in range(n_cols, n_nodes, n_cols):
        link_dirs_at_node[node, 0] = - 1
        link_dirs_at_node[node, 1] = - 1
        link_dirs_at_node[node, 2] = 0
        link_dirs_at_node[node, 3] = 1

    # Right nodes
    for node in range(2 * n_cols - 1, n_nodes, n_cols):
        link_dirs_at_node[node, 0] = 0
        link_dirs_at_node[node, 1] = - 1
        link_dirs_at_node[node, 2] = 1
        link_dirs_at_node[node, 3] = 1

    node = 0
    link_dirs_at_node[node, 0] = - 1
    link_dirs_at_node[node, 1] = - 1
    link_dirs_at_node[node, 2] = 0
    link_dirs_at_node[node, 3] = 0

    node = n_cols - 1
    link_dirs_at_node[node, 0] = 0
    link_dirs_at_node[node, 1] = - 1
    link_dirs_at_node[node, 2] = 1
    link_dirs_at_node[node, 3] = 0

    node = n_nodes - n_cols
    link_dirs_at_node[node, 0] = - 1
    link_dirs_at_node[node, 1] = 0
    link_dirs_at_node[node, 2] = 0
    link_dirs_at_node[node, 3] = 1

    node = n_nodes - 1
    link_dirs_at_node[node, 0] = 0
    link_dirs_at_node[node, 1] = 0
    link_dirs_at_node[node, 2] = 1
    link_dirs_at_node[node, 3] = 1



"""Calculates erosion rate as a function of the depth-slope product

Erosion rate = k_e * ((Tau**a - Tau_crit**a))

k_e = erodibility coefficient
Tau = bed shear stress 
    = density of fluid (rho) * gravitational acceleration (g) * water depths (h) * slopes (S)
Tau_crit = critical shear stress
a = positive exponent

Note this equation was presented in Tucker, G.T., 2004, Drainage basin 
sensitivityto tectonic and climatic forcing: Implications of a stochastic 
model for the role of entrainment and erosion thresholds, 
Earth Surface Processes and Landforms.

More generalized than other erosion components, as it doesn't require the 
upstream node order, links to flow receiver and flow receiver fields. Instead, 
takes in the water depth and slope fields on NODES calculated by the 
OverlandFlow class and erodes the landscape in response to the hydrograph
generted by that method.

As of right now, this component relies on the OverlandFlow component 
for stability. There are no stability criteria implemented in this class. 
To ensure model stability, use StreamPowerEroder or FastscapeEroder
components instead. 

.. codeauthor:: Jordan Adams

Examples
--------
>>> import numpy as np
>>> from landlab import RasterModelGrid
>>> from landlab.components import DepthSlopeProductErosion

Create a grid on which to calculate detachment ltd sediment transport.

>>> grid = RasterModelGrid((5, 5))

The grid will need some data to provide the detachment limited sediment
transport component. To check the names of the fields that provide input to
the detachment ltd transport component, use the *input_var_names* class
property.

Create fields of data for each of these input variables.

First create toopgraphy. This is a flat surface of elevation 10 m. 
>>> grid.at_node['topographic__elevation'] = np.ones(grid.number_of_nodes)
>>> grid.at_node['topographic__elevation'] *= 10.
>>> grid.at_node['topographic__elevation'] = np.array([
...      10., 10., 10., 10., 10., 
...      10., 10., 10., 10., 10., 
...      10., 10., 10., 10., 10., 
...      10., 10., 10., 10., 10., 
...      10., 10., 10., 10., 10.])

Now we'll add an arbitrary water depth field on top of that topography.
>>> grid.at_node['surface_water__depth'] = np.array([
...      5., 5., 5., 5., 5., 
...      4., 4., 4., 4., 4.,
...      3., 3., 3., 3., 3.,
...      2., 2., 2., 2., 2.,
...      1., 1., 1., 1., 1.])

Using the set topography, now we will calculate slopes on all nodes.

First calculating slopes on links
>>> grid.at_link['water_surface__slope'] = grid.calc_grad_at_link('surface_water__depth')

Now putting slopes on nodes

>>> grid['node']['water_surface__slope'] = (grid['link']['water_surface__slope'][grid.links_at_node] * grid.active_link_dirs_at_node).max(axis=1) # doctest: +NORMALIZE_WHITESPACE
>>> grid.at_node['water_surface__slope']
array([ 0.,  1.,  1.,  1.,  0., -0.,  1.,  1.,  1.,  0., -0.,  1.,  1.,
        1.,  0., -0.,  1.,  1.,  1.,  0.,  0.,  0.,  0.,  0.,  0.])


Instantiate the `DepthSlopeProductErosion` component to work on this grid, and
run it. In this simple case, we need to pass it a time step ('dt') and also
an erodibility factor ('k_e').

>>> dt = 1.
>>> dspe = DepthSlopeProductErosion(grid, k_e=0.00005)
>>> dspe.erode(dt=dt, slope='water_surface__slope')

Now we test to see how the topography changed as a function of the erosion
rate. First, we'll look at the erosion rate:
    
>>> dspe.dz   # doctest: +NORMALIZE_WHITESPACE 
array([ 0.    , -2.4525, -2.4525, -2.4525,  0.    ,  0.    , -1.962 ,
       -1.962 , -1.962 ,  0.    ,  0.    , -1.4715, -1.4715, -1.4715,
        0.    ,  0.    , -0.981 , -0.981 , -0.981 ,  0.    ,  0.    ,
        0.    ,  0.    ,  0.    ,  0.    ])

Now, our updated topography...
>>> grid.at_node['topographic__elevation'] # doctest: +NORMALIZE_WHITESPACE
array([ 10.    ,   7.5475,   7.5475,   7.5475,  10.    ,  10.    ,
         8.038 ,   8.038 ,   8.038 ,  10.    ,  10.    ,   8.5285,
         8.5285,   8.5285,  10.    ,  10.    ,   9.019 ,   9.019 ,
         9.019 ,  10.    ,  10.    ,  10.    ,  10.    ,  10.    ,  10.    ])
"""

from landlab import Component
import pylab
import numpy as np
from matplotlib import pyplot as plt
from landlab.field.scalar_data_fields import FieldError


class DepthSlopeProductErosion(Component):

    """Landlab component that simulates detachment-limited river erosion.

    This component calculates changes in elevation in response to vertical
    incision.
    """

    _name = 'DetachmentLtdErosion'

    _input_var_names = (
        'topographic__elevation',
        'topographic__slope',
        'surface_water__depth',
    )

    _output_var_names = (
        'topographic__elevation',
    )

    _var_units = {
        'topographic__elevation': 'm',
        'topographic__slope': '-',
        'surface_water__depth': 'm',
    }

    _var_mapping = {
        'topographic__elevation': 'node',
        'topographic__slope': 'node',
        'surface_water__depth': 'node',
    }

    _var_doc = {
        'topographic__elevation': 'Land surface topographic elevation',
        'topographic__slope': 'Slope of the land surface',
        'surface_water__depth': 'Depth of water on the surface',
    }

    def __init__(self, grid, k_e, fluid_density=1000., g=9.81, 
                 a_exp=1.0, tau_crit=0.0, uplift_rate=0.0, **kwds):
        """Calculate detachment limited erosion rate on nodes using the shear
        stress equation, solved using the depth slope product.

        Landlab component that generalizes the detachment limited erosion
        equation, primarily to be coupled to the the Landlab OverlandFlow
        component.

        This component adjusts topographic elevation and is contained in the
        landlab.components.detachment_ltd_erosion folder.

        Parameters
        ----------
        grid : RasterModelGrid
            A landlab grid.
        k_e : float
            Erodibility parameter, (m^(1+a_exp)*s^(2*a_exp-1)/kg^a_exp)
        fluid_density : float, optional
            Density of fluid, default set to water density of 1000 kg / m^3
        g : float, optional
            Acceleration due to gravity (m/s^2).
        a_exp : float, optional
            exponent on shear stress, positive, unitless
        tau_crit : float, optional
            threshold for sediment movement, (kg/m/s^2)
        uplift_rate : float, optional
            uplift rate applied to the topographic surface, m/s
        """
        super(DepthSlopeProductErosion, self).__init__(grid, **kwds)

        self.a = a_exp
        self.g = g
        self.rho = fluid_density
        self.E = self._grid.zeros(at='node')
        self.uplift_rate = uplift_rate
        self.tau_crit = tau_crit
        self.k_e = k_e

        self.dz = self._grid.zeros(at='node')

    def erode(self, dt, elevs='topographic__elevation', 
                          depth='surface_water__depth', 
                          slope='topographic__slope'):
            
        """Erode into grid topography.
        

        For one time step, this erodes into the grid topography using
        the water discharge and topographic slope.

        The grid field 'topographic__elevation' is altered each time step.

        Parameters
        ----------
        dt : float
            Time step.
        elevs : str, optional
            Name of the field that represents topographic elevation on nodes.
        depth : str, optional 
            Name of the field that represents water depths on nodes.
        slope : str, optional
            Name of the field that represent topographic slope on each node.
        """
        
        
        try:
            S = self._grid.at_node[slope]
        except FieldError:
            raise ValueError('Slope field is missing!')
            
        try:
            h = self._grid.at_node[depth]
        except FieldError:
            raise ValueError('Depth field is missing!')

        self.tau = self.rho * self.g * h * S
        
        greater_than_tc,  = np.where(self.tau >= self.tau_crit)
        less_than_tc,  = np.where(self.tau < self.tau_crit)
        
        self.E[less_than_tc] = 0.0
        
        self.E[greater_than_tc] = (self.k_e * 
                                 ((self.tau[greater_than_tc] ** self.a) - 
                                 (self.tau_crit ** self.a)))
        
        
        self.E[self.E < 0.0] = 0.0

        self.dz = (self.uplift_rate - self.E) * dt

        self._grid['node'][elevs] += self.dz

    def run_one_step(self, dt, elevs='topographic__elevation', 
                          depth='surface_water__depth', 
                          slope='topographic__slope'):
        
        self.erode(dt=dt, elevs=elevs, depth=depth, slope=slope)



"""
Utility functions that operate on landlab grids.
------------------------------------------------

"""


import numpy as np
from six.moves import range


def resolve_values_on_active_links(grid, active_link_values):
    """Resolve active-link values into x and y directions.

    Takes a set of values defined on active links, and returns those values
    resolved into the x and y directions.  Two link arrays are returned:
    x, then y.

    Parameters
    ----------
    grid : ModelGrid
        A ModelGrid.
    active_link_values : ndarray
        Values on active links.

    Returns
    -------
    tuple of ndarray
        Values resolved into x-component and y-component.
    """
    link_lengths = grid.length_of_link[grid.active_links]
    return (
        np.multiply(((grid.node_x[grid._activelink_tonode] -
                      grid.node_x[grid._activelink_fromnode]) /
                     link_lengths), active_link_values),
        np.multiply(((grid.node_y[grid._activelink_tonode] -
                      grid.node_y[grid._activelink_fromnode]) /
                     link_lengths), active_link_values))


def resolve_values_on_links(grid, link_values):
    """Resolve link values into x and y directions.

    Takes a set of values defined on active links, and returns those values
    resolved into the x and y directions.  Two link arrays are returned:
    x, then y.

    Parameters
    ----------
    grid : ModelGrid
        A ModelGrid.
    link_values : ndarray
        Values on links.

    Returns
    -------
    tuple of ndarray
        Values resolved into x-component and y-component.
    """
    return (
        np.multiply(((grid.node_x[grid.node_at_link_head] -
                      grid.node_x[grid.node_at_link_tail]) /
                     grid.length_of_link), link_values),
        np.multiply(((grid.node_y[grid.node_at_link_head] -
                      grid.node_y[grid.node_at_link_tail]) /
                     grid.length_of_link), link_values))



#! /usr/env/python
"""Read input variables for landlab components.

The Model Parameter Dictionary is a tool for numerical modelers to
easily read and access model parameters from a simple formatted
input (text) file. Each parameter has a KEY, which identifies the
parameter, and a VALUE, which can be a number or a string. A
ModelParameterDictionary object reads model parameters from an input
file to a Dictionary, and provides functions for the user to look up
particular parameters by key name.

The format of the input file looks like::

    >>> from six import StringIO
    >>> param_file = StringIO('''
    ... PI: the text "PI" is an example of a KEY
    ... 3.1416
    ... AVOGADROS_NUMBER: this is another
    ... 6.022e23
    ... FAVORITE_FRUIT: yet another
    ... mangoes
    ... NUMBER_OF_MANGO_WALKS: this one is an integer
    ... 4
    ... ALSO_LIKES_APPLES: this is a boolean
    ... true
    ... ''')

Example code that reads these parameters from a file called
*myinputs.txt*:

    >>> from landlab import ModelParameterDictionary
    >>> my_param_dict = ModelParameterDictionary()
    >>> my_param_dict.read_from_file(param_file)
    >>> pi = my_param_dict.read_float('PI')
    >>> avogado = my_param_dict.read_float('AVOGADROS_NUMBER')
    >>> fruit = my_param_dict.read_string('FAVORITE_FRUIT')
    >>> nmang = my_param_dict.read_int('NUMBER_OF_MANGO_WALKS')
    >>> apples_ok = my_param_dict.read_bool('ALSO_LIKES_APPLES')

As in Python, hash marks (#) denote comments. The rules are that each
key must have one and only one parameter value, and each value must
appear on a separate line immediately below the key line.

Also available are functions to read input parameters from the
command line (e.g., read_float_cmdline( 'PI' ) )
"""

# Licensing information:
#
# model_parameter_dictionary.py: reads formatted input from a text file
#   for use in specifying parameter values in numerical models.
#
# Copyright (C) 2011 Gregory E. Tucker
#
# Developer can be contacted at:
#   Cooperative Institute for Research in Environmental Sciences (CIRES)
#   University of Colorado Boulder
#   Campus Box 399
#   Boulder, CO 80309 USA
#   Phone: (+1) 303-492-6985
#   Email: gtucker@colorado.edu
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor,
# Boston, MA 02110-1301 USA.


import warnings
import six

import numpy as np


_VALID_TRUE_VALUES = set(['TRUE', '1', 1])
_VALID_FALSE_VALUES = set(['FALSE', '0', 0])
_VALID_BOOLEAN_VALUES = _VALID_TRUE_VALUES | _VALID_FALSE_VALUES


class Error(Exception):

    """Base class for exceptions raised from this module."""

    pass


class MissingKeyError(Error):

    """Error to indicate a missing parameter key.

    Raise this error if the parameter dictionary file does not contain a
    requested *key*.
    """

    def __init__(self, key):
        self._key = key

    def __str__(self):
        return self._key


class ParameterValueError(Error):

    """Error to indicate a bad parameter values.

    Raise this error if a parameter value given by *key* is not of the
    expected type.
    """

    def __init__(self, key, val, expected_type):
        self._key = key
        self._val = val
        self._type = expected_type

    def __str__(self):
        return '%s: %s is not of type %s' % (self._key, self._val, self._type)


def _to_bool(string):
    """Convert a string to a boolean."""
    upper = string.upper()
    if upper in _VALID_TRUE_VALUES:
        return True
    elif upper in _VALID_FALSE_VALUES:
        return False
    else:
        raise ValueError('invalid literal for _to_bool()')


_CONVERT_FROM_STR = {
    'float': float,
    'int': int,
    'str': str,
    'bool': _to_bool,
}


_VALID_VALUE_TYPES = set(_CONVERT_FROM_STR)


def _value_is_array(value):
    """Check if a parameter value is an array.

    Parameters
    ----------
    value : str
        Parameter value as a string.

    Returns
    -------
    bool
        ``True`` if the value is an array. Otherwise, ``False``.
    """
    return ',' in value


def _value_to_array(value):
    """Convert a string value to an array.

    Parameters
    ----------
    value : str
        Parameter value as a string.

    Raises
    ------
    ValueError
        If the string cannot be converted.

    Returns
    -------
    ndarray
        An array of values.
    """
    try:
        return np.array(value.split(','), np.int)
    except ValueError:
        return np.array(value.split(','), np.float)


def _value_to_numeric(value):
    """Convert a value string to a number.

    Parameters
    ----------
    value : str
        Parameter value as a string.

    Raises
    ------
    ValueError
        If the string cannot be converted.

    Returns
    -------
    number
        The value converted to ``int`` or ``float``.
    """
    try:
        return int(value)
    except ValueError:
        return float(value)


def _value_to_bool(value):
    """Convert a string value to a bool."""
    return value.upper() in _VALID_TRUE_VALUES


def _value_to_scalar(value):
    """Convert a string value to a scalar."""
    if value.upper() in _VALID_BOOLEAN_VALUES:
        return _value_to_bool(value)
    else:
        return _value_to_numeric(value)


class ModelParameterDictionary(dict):

    """Model parameter file as specified as key/value pairs.

    Reads model parameters from an input file to a dictionary
    and provides functions for the user to look up particular parameters
    by key name.

    If the keyword *auto_type* is True, then guess at the type for each value.
    Use *from_file* to read in a parameter file from a file-like object or a
    file with the given file name.

    Parameters
    ----------
    from_file : str or file_like, optional
        File from which to read parameters.
    auto_type : boolean, optional
        Try to guess parameter data types.

    Examples
    --------
    Create a file-like object that contains a model parameter dictionary.

    >>> from six import StringIO
    >>> test_file = StringIO('''
    ... INT_VAL:
    ... 1
    ... DBL_VAL:
    ... 1.2
    ... BOOL_VAL:
    ... true
    ... INT_ARRAY:
    ... 1,2,3
    ... DBL_ARRAY:
    ... 1.,2.,3.
    ... STR_VAL:
    ... landlab is awesome!
    ... ''')

    Create a ModelParameterDictionary, fill it with values from the
    parameter dictionary, and try to convert each value string to its
    intended type.

    >>> from landlab import ModelParameterDictionary
    >>> params = ModelParameterDictionary(auto_type=True, from_file=test_file)

    The returned ModelParameterDictionary can now be used just like a
    regular Python dictionary to get items, keys, etc.

    >>> sorted(params.keys())
    ['BOOL_VAL', 'DBL_ARRAY', 'DBL_VAL', 'INT_ARRAY', 'INT_VAL', 'STR_VAL']

    >>> params['INT_VAL']
    1
    >>> params['DBL_VAL']
    1.2
    >>> params['BOOL_VAL']
    True
    >>> params['STR_VAL']
    'landlab is awesome!'

    Lines containing commas are converted to numpy arrays. The type of the
    array is determined by the values.

    >>> isinstance(params['DBL_ARRAY'], np.ndarray)
    True
    >>> params['INT_ARRAY']
    array([1, 2, 3])
    >>> params['DBL_ARRAY']
    array([ 1.,  2.,  3.])
    """

    def __init__(self, from_file=None, auto_type=False):
        """Create a ModelParameterDictionary.

        Parameters
        ----------
        from_file : str or file_like, optional
            File from which to read parameters.
        auto_type : boolean, optional
            Try to guess parameter data types.
        """
        super(ModelParameterDictionary, self).__init__()

        self._auto_type = auto_type
        if from_file is not None:
            self.read_from_file(from_file)

    def read_from_file(self, param_file):
        """Read parameters for a file.

        Read and parse parameter dictionary information from a file or
        file-like object in *param_file*.

        The format of the parameter file should be like::

            # A comment line
            SOME_KEY: this the key for some parameter
            1.234

        In other words, the rules are:

        - Comments are preceded by hash characters
        - Each parameter has two consecutive lines, one for the key and one
          for the value
        - The key must be followed by a space, colon, or eol
        - The parameter can be numeric or text

        Parameters
        ----------
        param_file : str or file_like
            Name of parameter file (or file_like)
        """
        if isinstance(param_file, six.string_types):
            try:
                with open(param_file, 'r') as opened_file:
                    self._read_from_file_like(opened_file)
            except IOError:
                raise
        else:
            self._read_from_file_like(param_file)

    @staticmethod
    def _get_stripped_lines(param_file):
        """Strip whitespace for the lines of a parameter file.

        Strip lines from the iterable, *param_file*. Ignore lines, that upon
        being stripped, are either blank or only contain a comment.
        Comments are lines that contain a hash preceeded only by whitespace.

        Returns a list of the stripped lines.

        Parameters
        ----------
        param_file : file_like
            A parameter file.

        Returns
        -------
        list
            The lines of the parameter file with leading and trailing
            whitespace stripped off.
        """
        stripped_line_list = []
        for line in param_file:
            line = line.strip()   # strip leading spaces
            if len(line) > 0 and line[0] != '#':
                stripped_line_list.append(line)
        return stripped_line_list

    def _read_from_file_like(self, param_file):
        """Read parameters from a file-like object.

        Read parameters from the file-like object, *param_file*. In fact,
        *param_file* really only needs to be an iterable of strings.

        Parameters
        ----------
        param_file : iterable of str
            List of strings as lines from a parameter file.
        """
        stripped_line_list = self._get_stripped_lines(param_file)

        iskey = True
        for line in stripped_line_list:
            if iskey:
                # Strip out everything after the first space or colon
                first_colon = line.find(':')
                if first_colon == -1:
                    first_colon = len(line)
                first_space = line.find(' ')
                if first_space == -1:
                    first_space = len(line)
                last_key_char = min(first_colon, first_space)
                last_key = line[0:last_key_char]  # remember the key
                iskey = False
            else:
                if self._auto_type:
                    self[last_key] = self._auto_type_value(line)
                else:
                    self[last_key] = line
                iskey = True

    @staticmethod
    def _auto_type_value(line):
        """Guess the data type of a parameter value.

        Read a value from a string and try to guess its type. If the line
        contains any commas, try to convert it to a numpy array of ints and,
        if that doesn't work, an array of floats, otherwise it's just a
        string.

        If there are no commas, the order of types is ``bool``, ``int``,
        ``float``, and ``str``.

        Parameters
        ----------
        line : str
            A line from a parameter file.

        Returns
        -------
        The value converted to the guessed data type.
        """
        try:
            if _value_is_array(line):
                return _value_to_array(line)
            else:
                return _value_to_scalar(line)
        except ValueError:
            return line

    def params(self):
        """List of all the parameters names in the parameter dictionary.

        Returns
        -------
        list of str
            The names of parameters in the file.
        """
        return self.keys()

    def get(self, key, *args, **kwds):
        """get(key, [default], ptype=str)
        Get a value by key name.

        Get a value from a model parameter dictionary. Use the *ptype*
        keyword to convert the value to a given type. *ptype* is a function
        that converts the retreived value to the desired value. If a second
        argument after *key* is provided, use it as a default in case *key*
        is not contained in the ModelParameterDictionary.

        Parameters
        ----------
        key : str
            A parameter name.
        default : str or number, optional
            A default value if the key is missing.
        ptype : str, optional
            The data type of the paramter.

        Examples
        --------
        >>> from six import StringIO
        >>> from landlab import ModelParameterDictionary
        >>> params = ModelParameterDictionary(StringIO(
        ... '''
        ... MY_INT:
        ... 1
        ... '''))
        >>> params.get('MY_INT')
        '1'
        >>> params.get('MY_INT', ptype=int)
        1
        >>> params.get('MY_MISSING_INT', 2, ptype=int)
        2

        Be careful when dealing with booleans. If you want to be returned
        a boolean value, *DO NOT* set the *ptype* keyword to the builtin
        *bool*. This will not work as the Python *bool* function does not
        convert strings to booleans as you might expect. For example::

            >>> bool('True')
            True
            >>> bool('False')
            True

        If you would like to get a boolean, use ``ptype='bool'``.

        .. note:: Use ``ptype='bool'`` not ``ptype=bool``.

            If you use *bool* to convert a string the returned boolean will
            be ``True`` for *any* non-empty string. This is just how the
            Python built-in ``bool`` works::

                >>> bool('0')
                True
                >>> bool('1')
                True
                >>> bool('')
                False

        >>> from six import StringIO
        >>> params = ModelParameterDictionary(StringIO(
        ... '''
        ... MY_BOOL:
        ... false
        ... '''))
        >>> params.get('MY_BOOL')
        'false'
        >>> params.get('MY_BOOL', ptype='bool')
        False
        """
        ptype = kwds.pop('ptype', str)
        assert len(kwds) == 0

        if ptype is bool:
            warnings.warn(
                "Using bool function to convert value. This probably isn't "
                "what you want. Use ptype='bool' instead.",
                RuntimeWarning)

        value = super(ModelParameterDictionary, self).get(key, *args)
        if value is None:
            raise MissingKeyError(key)

        if isinstance(ptype, str):
            try:
                converter = _CONVERT_FROM_STR[ptype]
            except KeyError:
                converter = str
        else:
            converter = ptype

        try:
            typed_value = converter(value)
        except ValueError:
            raise ParameterValueError(key, value, ptype)
        else:
            return typed_value

    def read_int(self, key, *args):
        """read_int(key, [default])
        Locate *key* in the input file and return it as an integer.

        Parameters
        ----------
        key : str
            The name of a parameter.
        default : int
            The default value if the key is missing.

        Returns
        -------
        int
            The value type cast to an ``int``.

        Raises
        ------
        MissingKeyError
            If *key* isn't in the dictionary or if its value is not an integer.
        ParameterValueError
            If the value is not an integer.

        Examples
        --------
        >>> from six import StringIO
        >>> from landlab import ModelParameterDictionary
        >>> params = ModelParameterDictionary(StringIO(
        ... '''
        ... MY_INT:
        ... 1
        ... '''))
        >>> params.read_int('MY_INT')
        1
        """
        return self.get(key, *args, ptype=int)

    def read_float(self, key):
        """read_float(key, [default])
        Locate *key* in the input file and return it as a float.

        Parameters
        ----------
        key : str
            The name of a parameter.
        default : float
            The default value if the key is missing.

        Returns
        -------
        float
            The value type cast to a ``float``.

        Raises
        ------
        MissingKeyError
            If *key* isn't in the dictionary or if its value is not an integer.
        ParameterValueError
            If the value is not a float.

        Examples
        --------
        >>> from __future__ import print_function
        >>> from six import StringIO
        >>> from landlab import ModelParameterDictionary
        >>> params = ModelParameterDictionary(StringIO(
        ... '''
        ... MY_FLOAT:
        ... 3.14
        ... '''))
        >>> print('%.2f' % params.read_float('MY_FLOAT'))
        3.14
        """
        try:
            my_float = float(self[key])
        except KeyError:
            raise MissingKeyError(key)
        except ValueError:
            raise ParameterValueError(key, self[key], 'float')
        else:
            return my_float

    def read_string(self, key):
        """Locate *key* in the input file and return it as a string.

        Parameters
        ----------
        key : str
            The name of a parameter.
        default : str
            The default value if the key is missing.

        Returns
        -------
        str
            The value type cast to a ``str``.

        Raises
        ------
        MissingKeyError
            If *key* isn't in the dictionary or if its value is not an integer.

        Examples
        --------
        >>> from six import StringIO
        >>> from landlab import ModelParameterDictionary
        >>> params = ModelParameterDictionary(StringIO(
        ... '''
        ... MY_STRING:
        ... landlab
        ... '''))
        >>> params.read_string('MY_STRING')
        'landlab'
        """
        try:
            my_value = self[key]
        except KeyError:
            raise MissingKeyError(key)
        return str(my_value)

    def read_bool(self, key):
        """Locate *key* in the input file and return it as a boolean.

        Parameters
        ----------
        key : str
            The name of a parameter.
        default : bool
            The default value if the key is missing.

        Returns
        -------
        bool
            The value type cast to a ``bool``.

        Raises
        ------
        MissingKeyError
            If *key* isn't in the dictionary or if its value is not an integer.
        ParameterValueError
            If the value is not a boolean.

        Examples
        --------
        >>> from six import StringIO
        >>> from landlab import ModelParameterDictionary
        >>> params = ModelParameterDictionary(StringIO(
        ... '''
        ... MY_BOOL:
        ... true
        ... '''))
        >>> params.read_bool('MY_BOOL')
        True
        """
        try:
            my_value = self[key]
        except KeyError:
            raise MissingKeyError(key)

        if my_value.upper() in _VALID_TRUE_VALUES:
            return True
        elif my_value.upper() in _VALID_FALSE_VALUES:
            return False
        else:
            raise ParameterValueError(key, my_value, 'boolean')

    def read_int_cmdline(self, key):
        """Read an integer from the command line.

        Read an integer from the command line and use it as a value for
        *key* in the dictonary.

        Parameters
        ----------
        key : str
            The name of a parameter.

        Returns
        -------
        int
            The parameter value.

        Raises
        ------
        ParameterValueError
            If the value is not an int.
        """
        my_value = raw_input(key + ': ')
        self[key] = my_value
        if not isinstance(my_value, int):
            raise ParameterValueError(key, my_value, 'int')
        return my_value

    def read_float_cmdline(self, key):
        """Read a float from the command line.

        Read a float from the command line and use it as a value for
        *key* in the dictonary.

        Parameters
        ----------
        key : str
            The name of a parameter.

        Returns
        -------
        float
            The parameter value.

        Raises
        ------
        ParameterValueError
            If the value is not an float.
        """
        my_value = raw_input(key + ': ')
        self[key] = my_value
        try:
            my_float = float(my_value)
        except ValueError:
            raise ParameterValueError(key, my_value, 'float')
        else:
            return my_float

    def read_string_cmdline(self, key):
        """Read a string from the command line.

        Read a string from the command line and use it as a value for
        *key* in the dictonary.

        Parameters
        ----------
        key : str
            The name of a parameter.

        Returns
        -------
        str
            The parameter value.
        """
        my_str = raw_input(key + ': ')
        self[key] = my_str
        return my_str

    @staticmethod
    def read_bool_cmdline(key):
        """Read a boolean from the command line.

        Read a boolean from the command line and use it as a value for
        *key* in the dictonary.

        Parameters
        ----------
        key : str
            The name of a parameter.

        Returns
        -------
        bool
            The parameter value.
        """
        my_value = raw_input(key + ': ')
        if my_value.upper() in _VALID_TRUE_VALUES:
            return True
        elif my_value.upper() in _VALID_FALSE_VALUES:
            return False
        else:
            raise ParameterValueError(key, my_value, 'boolean')



#! /usr/bin/env python
"""
Unit tests for landlab.components.flexure.flexure
"""
import pytest

import numpy as np


(_SHAPE, _SPACING, _ORIGIN) = ((20, 20), (10e3, 10e3), (0., 0.))
_ARGS = (_SHAPE, _SPACING, _ORIGIN)


def test_name(flex):
    assert flex.name == "Flexure"


def test_input_var_names(flex):
    assert flex.input_var_names == ("lithosphere__overlying_pressure_increment",)


def test_output_var_names(flex):
    assert flex.output_var_names == ("lithosphere_surface__elevation_increment",)


def test_var_units(flex):
    assert set(flex.input_var_names) | set(flex.output_var_names) == set(
        dict(flex.units).keys()
    )

    assert flex.var_units("lithosphere_surface__elevation_increment") == "m"
    assert flex.var_units("lithosphere__overlying_pressure_increment") == "Pa"


def test_grid_shape(flex):
    assert flex.grid.number_of_node_rows == _SHAPE[0]
    assert flex.grid.number_of_node_columns == _SHAPE[1]


def test_grid_x_extent(flex):
    assert flex.grid.extent[1] == (_SHAPE[1] - 1) * _SPACING[1]


def test_grid_y_extent(flex):
    assert flex.grid.extent[0] == (_SHAPE[0] - 1) * _SPACING[0]


def test_field_getters(flex):
    for name in flex.grid["node"]:
        field = flex.grid["node"][name]
        assert isinstance(field, np.ndarray)
        assert field.shape == (
            flex.grid.number_of_node_rows * flex.grid.number_of_node_columns,
        )

    with pytest.raises(KeyError):
        flex.grid["not_a_var_name"]


def test_field_initialized_to_zero(flex):
    for name in flex.grid["node"]:
        field = flex.grid["node"][name]
        assert np.all(field == 0.)



#! /usr/bin/env python
"""Unit tests for landlab.io.netcdf module."""

import os

import pytest

from landlab.io.netcdf import read_netcdf, WITH_NETCDF4


_TEST_DATA_DIR = os.path.join(os.path.dirname(__file__), 'data')


def test_read_netcdf3_64bit():
    """Test read_netcdf for with 64-bit netcdf3 format."""
    grid = read_netcdf(os.path.join(_TEST_DATA_DIR, 'test-netcdf3-64bit.nc'))
    assert grid.shape == (4, 3)


@pytest.mark.skipif(not WITH_NETCDF4, reason='netCDF4 package not installed')
def test_read_netcdf4():
    """Test read_netcdf with netcdf4 format."""
    grid = read_netcdf(os.path.join(_TEST_DATA_DIR, 'test-netcdf4.nc'))
    assert grid.shape == (4, 3)
